from __future__ import annotations

import multiprocessing.managers
import time
from collections.abc import Mapping
from typing import Optional

import tblib.pickling_support

from caqtus.device import DeviceName, DeviceConfiguration
from caqtus.session import ExperimentSessionMaker, PureSequencePath
from caqtus.types.parameter import ParameterNamespace
from .manager import ExperimentManager, Procedure, LocalExperimentManager
from ..device_manager_extension import DeviceManagerExtensionProtocol
from ..sequence_runner import ShotRetryConfig

experiment_manager: Optional[LocalExperimentManager] = None


class ExperimentManagerProxy(ExperimentManager, multiprocessing.managers.BaseProxy):
    _exposed_ = ("create_procedure", "interrupt_running_procedure")
    _method_to_typeid_ = {
        "create_procedure": "ProcedureProxy",
    }

    def create_procedure(
        self, procedure_name: str, acquisition_timeout: Optional[float] = None
    ) -> ProcedureProxy:
        return self._callmethod("create_procedure", (procedure_name,))  # type: ignore

    def interrupt_running_procedure(self) -> bool:
        return self._callmethod("interrupt_running_procedure", ())  # type: ignore

    def __repr__(self):
        return f"<ExperimentManagerProxy at {hex(id(self))}>"


class ProcedureProxy(Procedure, multiprocessing.managers.BaseProxy):
    """Proxy for a procedure running in a different process.

    This object behaves like a :class:`Procedure` object and should be used like it, but
    it forwards all method calls to an actual procedure object running in a different
    process.
    """

    _exposed_ = (
        "__enter__",
        "__exit__",
        "is_active",
        "is_running_sequence",
        "sequences",
        "exception",
        "start_sequence",
        "interrupt_sequence",
        "run_sequence",
        "__str__",
    )
    _method_to_typeid_ = {"__enter__": "ProcedureProxy"}

    def __enter__(self):
        return self._callmethod("__enter__", ())

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self._callmethod("__exit__", (None, exc_val, None))

    def __repr__(self):
        return f"<{self.__class__.__name__}('{self}') at {hex(id(self))}>"

    def __str__(self):
        return self._callmethod("__str__", ())

    def is_active(self) -> bool:
        return self._callmethod("is_active", ())  # type: ignore

    def is_running_sequence(self) -> bool:
        return self._callmethod("is_running_sequence", ())  # type: ignore

    def sequences(self) -> list[PureSequencePath]:
        return self._callmethod("sequences", ())  # type: ignore

    def exception(self) -> Optional[Exception]:
        while self.is_running_sequence():
            time.sleep(10e-3)
        return self._callmethod("exception", ())  # type: ignore

    def start_sequence(
        self,
        sequence: PureSequencePath,
        global_parameters: Optional[ParameterNamespace] = None,
        device_configurations: Optional[
            Mapping[DeviceName, DeviceConfiguration]
        ] = None,
    ) -> None:
        return self._callmethod(
            "start_sequence",
            (sequence, global_parameters, device_configurations),
        )

    def interrupt_sequence(self) -> bool:
        return self._callmethod("interrupt_sequence", ())  # type: ignore

    def run_sequence(
        self,
        sequence: PureSequencePath,
        global_parameters: Optional[ParameterNamespace] = None,
        device_configurations: Optional[
            Mapping[DeviceName, DeviceConfiguration]
        ] = None,
    ) -> None:
        # Here we can't just call the remote method `run_sequence`.
        # This is because this method takes a long time to return, since it waits for
        # the sequence to finish.
        # This means that we can't even raise KeyboardInterrupt while waiting for the
        # method to return.
        self.start_sequence(sequence, global_parameters, device_configurations)
        while self.is_running_sequence():
            time.sleep(10e-3)
        if exception := self.exception():
            raise exception


class _MultiprocessingServerManager(multiprocessing.managers.BaseManager):
    pass


def _get_experiment_manager() -> LocalExperimentManager:
    if experiment_manager is None:
        raise RuntimeError("Experiment manager not initialized")
    return experiment_manager


def _enter_experiment_manager() -> None:
    experiment_manager.__enter__()


def _exit_experiment_manager(exc_value) -> None:
    experiment_manager.__exit__(type(exc_value), exc_value, exc_value.__traceback__)


def _create_experiment_manager(
    session_maker: ExperimentSessionMaker,
    device_manager_extension: DeviceManagerExtensionProtocol,
    shot_retry_config: Optional[ShotRetryConfig] = None,
) -> None:
    tblib.pickling_support.install()
    global experiment_manager
    experiment_manager = LocalExperimentManager(
        session_maker=session_maker,
        shot_retry_config=shot_retry_config,
        device_manager_extension=device_manager_extension,
    )


_MultiprocessingServerManager.register(
    "create_experiment_manager", _create_experiment_manager, ExperimentManagerProxy
)
_MultiprocessingServerManager.register(
    "get_experiment_manager", _get_experiment_manager, ExperimentManagerProxy
)
_MultiprocessingServerManager.register(
    "enter_experiment_manager", _enter_experiment_manager
)
_MultiprocessingServerManager.register(
    "exit_experiment_manager", _exit_experiment_manager
)
_MultiprocessingServerManager.register("ProcedureProxy", None, ProcedureProxy)


class RemoteExperimentManagerServer:
    session_maker: Optional[ExperimentSessionMaker] = None

    def __init__(
        self,
        address: tuple[str, int],
        authkey: bytes,
        session_maker: ExperimentSessionMaker,
        device_manager_extension: DeviceManagerExtensionProtocol,
        shot_retry_config: Optional[ShotRetryConfig] = None,
    ):
        self._session_maker = session_maker
        self._multiprocessing_manager = _MultiprocessingServerManager(
            address=address, authkey=authkey
        )
        self._shot_retry_config = shot_retry_config
        self._shot_retry_config = shot_retry_config
        self._device_manager_extension = device_manager_extension

    def __enter__(self):
        self._multiprocessing_manager.start()
        self._multiprocessing_manager.create_experiment_manager(
            self._session_maker,
            self._device_manager_extension,
            self._shot_retry_config,
        )
        self._multiprocessing_manager.enter_experiment_manager()
        return self

    @staticmethod
    def serve_forever():
        while True:
            time.sleep(100e-3)

    def __exit__(self, exc_type, exc_value, traceback):
        self._multiprocessing_manager.exit_experiment_manager(exc_value)
        return self._multiprocessing_manager.__exit__(exc_type, exc_value, traceback)


class _MultiprocessingClientManager(multiprocessing.managers.BaseManager):
    pass


_MultiprocessingClientManager.register(
    "get_experiment_manager", None, ExperimentManagerProxy
)

_MultiprocessingClientManager.register("ProcedureProxy", None, ProcedureProxy)


class RemoteExperimentManagerClient:
    def __init__(self, address: tuple[str, int], authkey: bytes):
        self._multiprocessing_manager = _MultiprocessingClientManager(
            address=address, authkey=authkey
        )
        self._multiprocessing_manager.connect()

    def get_experiment_manager(self) -> ExperimentManagerProxy:
        return self._multiprocessing_manager.get_experiment_manager()



from collections.abc import Set
from typing import Optional, assert_never

from PySide6.QtCore import (
    QModelIndex,
    Qt,
    QAbstractItemModel,
    QObject,
    QEvent,
    QStringListModel,
)
from PySide6.QtGui import (
    QTextDocument,
    QPalette,
    QFocusEvent,
)
from PySide6.QtWidgets import (
    QWidget,
    QStyleOptionViewItem,
    QHBoxLayout,
    QLabel,
    QApplication,
    QSpinBox,
    QCompleter,
)

from caqtus.types.expression import Expression
from caqtus.types.variable_name import (
    DottedVariableName,
    InvalidVariableNameError,
)
from .steps_model import (
    StepData,
    VariableDeclarationData,
    LinspaceLoopData,
    ArrangeLoopData,
)
from ..._qt_util import AutoResizeLineEdit
from ....qtutil import HTMLItemDelegate

NAME_COLOR = "#AA4926"
VALUE_COLOR = "#6897BB"
HIGHLIGHT_COLOR = "#cc7832"


class StepDelegate(HTMLItemDelegate):
    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self.doc = QTextDocument(self)
        self._completer = QCompleter(self)

    def createEditor(
        self, parent: QWidget, option: QStyleOptionViewItem, index: QModelIndex
    ) -> QWidget:
        value = index.data(role=Qt.ItemDataRole.EditRole)
        assert isinstance(value, StepData)
        if isinstance(value, VariableDeclarationData):
            editor = VariableDeclarationEditor(parent, option.font)
        elif isinstance(value, LinspaceLoopData):
            editor = LinspaceLoopEditor(parent, option.font)
        elif isinstance(value, ArrangeLoopData):
            editor = ArrangeLoopEditor(parent, option.font)
        else:
            assert_never(value)
        editor.set_name_completer(self._completer)
        return editor

    def setEditorData(self, editor: QWidget, index: QModelIndex):
        data = index.data(role=Qt.ItemDataRole.EditRole)
        assert isinstance(data, StepData)
        match data:
            case VariableDeclarationData() as declaration:
                assert isinstance(editor, VariableDeclarationEditor)
                editor.set_step_data(declaration)
            case LinspaceLoopData() as loop:
                assert isinstance(editor, LinspaceLoopEditor)
                editor.set_step_data(loop)
            case ArrangeLoopData():
                assert isinstance(editor, ArrangeLoopEditor)
                editor.set_step_data(data)
            case _:
                raise ValueError(f"Can't set editor data for {data}")

    def updateEditorGeometry(
        self, editor: QWidget, option: QStyleOptionViewItem, index: QModelIndex
    ):
        geometry = option.rect
        editor.setGeometry(geometry)

    def setModelData(
        self, editor: QWidget, model: QAbstractItemModel, index: QModelIndex
    ) -> None:
        previous_data = index.data(role=Qt.ItemDataRole.EditRole)
        assert isinstance(previous_data, StepData)
        match previous_data:
            case VariableDeclarationData():
                assert isinstance(editor, VariableDeclarationEditor)
                try:
                    new_data = editor.get_step_data()
                except InvalidVariableNameError:
                    return
                else:
                    model.setData(index, new_data, Qt.ItemDataRole.EditRole)
            case LinspaceLoopData():
                assert isinstance(editor, LinspaceLoopEditor)
                try:
                    new_data = editor.get_step_data()
                except InvalidVariableNameError:
                    return
                else:
                    model.setData(index, new_data, Qt.ItemDataRole.EditRole)
            case ArrangeLoopData():
                assert isinstance(editor, ArrangeLoopEditor)
                try:
                    new_data = editor.get_step_data()
                except InvalidVariableNameError:
                    return
                else:
                    model.setData(index, new_data, Qt.ItemDataRole.EditRole)
            case _:
                assert_never(previous_data)

    def set_available_names(self, names: Set[DottedVariableName]) -> None:
        available_names = {str(name) for name in names}
        self._completer.setModel(QStringListModel(available_names))


class CompoundWidget(QWidget):
    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)
        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        self.setLayout(layout)
        self._widgets = []

    def add_widget(self, widget: QWidget):
        self.layout().addWidget(widget)
        widget.installEventFilter(self)
        self._widgets.append(widget)

    def eventFilter(self, watched: QObject, event: QEvent) -> bool:
        watched_is_child = watched in self._widgets
        if watched_is_child and event.type() == QEvent.Type.FocusOut:
            if QApplication.focusWidget() not in self.findChildren(QWidget):
                QApplication.postEvent(self, QFocusEvent(event.type()))
            return False

        return super().eventFilter(watched, event)


class LinspaceLoopEditor(CompoundWidget):
    def __init__(self, parent, font):
        super().__init__(parent)
        self.setFont(font)
        for_label = QLabel("for ", self)
        for_label.setAttribute(Qt.WA_TranslucentBackground, True)
        for_label.setStyleSheet(f"color: {HIGHLIGHT_COLOR}")
        self.add_widget(for_label)
        self.name_editor = AutoResizeLineEdit(self)
        self.name_editor.setStyleSheet(f"color: {NAME_COLOR}")
        self.add_widget(self.name_editor)
        equal_label = QLabel(" = ", self)
        equal_label.setAttribute(Qt.WA_TranslucentBackground, True)
        self.add_widget(equal_label)
        self.start_editor = AutoResizeLineEdit(self)
        self.start_editor.setStyleSheet(f"color: {VALUE_COLOR}")
        self.add_widget(self.start_editor)
        to_label = QLabel(" to ", self)
        to_label.setAttribute(Qt.WA_TranslucentBackground, True)
        to_label.setStyleSheet(f"color: {HIGHLIGHT_COLOR}")
        self.add_widget(to_label)
        self.stop_editor = AutoResizeLineEdit(self)
        self.stop_editor.setStyleSheet(f"color: {VALUE_COLOR}")
        self.add_widget(self.stop_editor)
        with_label = QLabel(" with ", self)
        with_label.setAttribute(Qt.WA_TranslucentBackground, True)
        with_label.setStyleSheet(f"color: {HIGHLIGHT_COLOR}")
        self.add_widget(with_label)
        self.num_editor = QSpinBox(self)
        self.num_editor.setStyleSheet(f"color: {VALUE_COLOR}")
        self.num_editor.setRange(0, 9999)
        self.add_widget(self.num_editor)
        steps_label = QLabel(" steps:", self)
        steps_label.setAttribute(Qt.WA_TranslucentBackground, True)
        steps_label.setStyleSheet(f"color: {HIGHLIGHT_COLOR}")
        self.add_widget(steps_label)
        self.layout().addStretch(1)

        palette = self.palette()
        palette.setColor(QPalette.ColorRole.Window, Qt.GlobalColor.black)
        self.setAutoFillBackground(True)
        self.setPalette(palette)

    def set_step_data(self, loop: LinspaceLoopData) -> None:
        self.name_editor.setText(str(loop.variable))
        self.start_editor.setText(str(loop.start))
        self.stop_editor.setText(str(loop.stop))
        self.num_editor.setValue(loop.num)

    def get_step_data(self) -> LinspaceLoopData:
        return LinspaceLoopData(
            variable=DottedVariableName(self.name_editor.text()),
            start=Expression(self.start_editor.text()),
            stop=Expression(self.stop_editor.text()),
            num=self.num_editor.value(),
        )

    def set_name_completer(self, completer: QCompleter) -> None:
        self.name_editor.setCompleter(completer)


class VariableDeclarationEditor(CompoundWidget):
    def __init__(self, parent, font):
        super().__init__(parent)
        self.setFont(font)
        self.name_editor = AutoResizeLineEdit(self)
        self.add_widget(self.name_editor)
        label = QLabel(" = ", self)
        label.setAttribute(Qt.WA_TranslucentBackground, True)
        self.add_widget(label)
        self.value_editor = AutoResizeLineEdit(self)
        self.add_widget(self.value_editor)
        self.layout().addStretch(1)

        palette = self.palette()
        palette.setColor(QPalette.ColorRole.Window, Qt.GlobalColor.black)
        self.setAutoFillBackground(True)
        self.setPalette(palette)
        self.name_editor.setStyleSheet(f"color: {NAME_COLOR}")
        self.value_editor.setStyleSheet(f"color: {VALUE_COLOR}")
        self.name_editor.setPlaceholderText("Parameter name")
        self.value_editor.setPlaceholderText("Parameter value")

    def set_step_data(self, declaration: VariableDeclarationData) -> None:
        self.name_editor.setText(str(declaration.variable))
        self.value_editor.setText(str(declaration.value))

    def get_step_data(self) -> VariableDeclarationData:
        return VariableDeclarationData(
            variable=DottedVariableName(self.name_editor.text()),
            value=Expression(self.value_editor.text()),
        )

    def set_name_completer(self, completer: QCompleter) -> None:
        self.name_editor.setCompleter(completer)


class ArrangeLoopEditor(CompoundWidget):
    def __init__(self, parent, font):
        super().__init__(parent)
        self.setFont(font)
        for_label = QLabel("for ", self)
        for_label.setAttribute(Qt.WA_TranslucentBackground, True)
        for_label.setStyleSheet(f"color: {HIGHLIGHT_COLOR}")
        self.add_widget(for_label)
        self.name_editor = AutoResizeLineEdit(self)
        self.name_editor.setStyleSheet(f"color: {NAME_COLOR}")
        self.add_widget(self.name_editor)
        equal_label = QLabel(" = ", self)
        equal_label.setAttribute(Qt.WA_TranslucentBackground, True)
        self.add_widget(equal_label)
        self.start_editor = AutoResizeLineEdit(self)
        self.start_editor.setStyleSheet(f"color: {VALUE_COLOR}")
        self.add_widget(self.start_editor)
        to_label = QLabel(" to ", self)
        to_label.setAttribute(Qt.WA_TranslucentBackground, True)
        to_label.setStyleSheet(f"color: {HIGHLIGHT_COLOR}")
        self.add_widget(to_label)
        self.stop_editor = AutoResizeLineEdit(self)
        self.stop_editor.setStyleSheet(f"color: {VALUE_COLOR}")
        self.add_widget(self.stop_editor)
        with_label = QLabel(" with ", self)
        with_label.setAttribute(Qt.WA_TranslucentBackground, True)
        with_label.setStyleSheet(f"color: {HIGHLIGHT_COLOR}")
        self.add_widget(with_label)
        self.step_editor = AutoResizeLineEdit(self)
        self.step_editor.setStyleSheet(f"color: {VALUE_COLOR}")
        self.add_widget(self.step_editor)
        spacing_label = QLabel(" spacing:", self)
        spacing_label.setAttribute(Qt.WA_TranslucentBackground, True)
        spacing_label.setStyleSheet(f"color: {HIGHLIGHT_COLOR}")
        self.add_widget(spacing_label)
        self.layout().addStretch(1)

        palette = self.palette()
        palette.setColor(QPalette.ColorRole.Window, Qt.GlobalColor.black)
        self.setAutoFillBackground(True)
        self.setPalette(palette)

    def set_step_data(self, loop: ArrangeLoopData) -> None:
        self.name_editor.setText(str(loop.variable))
        self.start_editor.setText(str(loop.start))
        self.stop_editor.setText(str(loop.stop))
        self.step_editor.setText(str(loop.step))

    def get_step_data(self) -> ArrangeLoopData:
        return ArrangeLoopData(
            variable=DottedVariableName(self.name_editor.text()),
            start=Expression(self.start_editor.text()),
            stop=Expression(self.stop_editor.text()),
            step=Expression(self.step_editor.text()),
        )

    def set_name_completer(self, completer: QCompleter) -> None:
        self.name_editor.setCompleter(completer)



import contextlib
from collections.abc import Mapping, AsyncGenerator, Callable
from typing import TypeVar, Any

from caqtus.device import DeviceName, Device, DeviceConfiguration
from caqtus.device.remote import DeviceProxy, RPCConfiguration
from caqtus.device.remote.rpc import RPCClient
from caqtus.experiment_control.device_manager_extension import (
    DeviceManagerExtensionProtocol,
)
from caqtus.formatter import fmt
from caqtus.types.recoverable_exceptions import ConnectionFailedError
from ._async_utils import task_group_with_error_message


@contextlib.asynccontextmanager
async def create_devices(
    initialization_parameters: Mapping[DeviceName, Mapping[str, Any]],
    device_configs: Mapping[DeviceName, DeviceConfiguration],
    device_types: Mapping[DeviceName, Callable[..., Device]],
    device_manager_extension: DeviceManagerExtensionProtocol,
) -> AsyncGenerator[dict[DeviceName, DeviceProxy], None]:
    device_server_configs = {}
    device_to_server = {}
    for device_name, device_config in device_configs.items():
        remote_server = device_config.remote_server
        if remote_server is not None:
            device_server_configs[remote_server] = (
                device_manager_extension.get_device_server_config(remote_server)
            )
            device_to_server[device_name] = remote_server

    async with create_rpc_clients(
        device_to_server, device_server_configs
    ) as rpc_clients:
        uninitialized_proxies = {}
        for device_name, init_params in initialization_parameters.items():
            device_config = device_configs[device_name]
            device_type = device_types[device_name]
            if device_config.remote_server is None:
                raise NotImplementedError
            client = rpc_clients[device_name]
            proxy_type = device_manager_extension.get_proxy_type(device_config)
            device_proxy = proxy_type(client, device_type, **init_params)
            uninitialized_proxies[device_name] = device_proxy
        async with context_group(uninitialized_proxies) as devices:
            yield devices


@contextlib.asynccontextmanager
async def create_rpc_clients(
    device_servers: Mapping[DeviceName, str],
    device_server_configs: Mapping[str, RPCConfiguration],
) -> AsyncGenerator[dict[DeviceName, RPCClient], None]:
    clients: dict[DeviceName, RPCClient] = {}
    async with contextlib.AsyncExitStack() as stack:
        for device_name, server in device_servers.items():
            config = device_server_configs[server]
            client = RPCClient(config.host, config.port)
            try:
                await stack.enter_async_context(client)
            except OSError as e:
                raise ConnectionFailedError(
                    fmt(
                        "Failed to connect to {:device server} for {:device}",
                        server,
                        device_name,
                    )
                ) from e
            clients[device_name] = client
        yield clients


T = TypeVar("T")


async def enter_and_push_device(
    stack: contextlib.AsyncExitStack,
    device: DeviceName,
    cm: contextlib.AbstractAsyncContextManager[T],
    results: dict,
):
    try:
        value = await stack.enter_async_context(cm)
    except Exception as e:
        raise RuntimeError(fmt("Failed to initialize {:device}", device)) from e
    results[device] = value


@contextlib.asynccontextmanager
async def context_group(
    cms: Mapping[str, contextlib.AbstractAsyncContextManager[T]]
) -> AsyncGenerator[Mapping[str, T], None]:
    results = {}
    async with contextlib.AsyncExitStack() as stack:
        async with task_group_with_error_message(
            "Errors occurred while initializing devices"
        ) as tg:
            for key, cm in cms.items():
                tg.start_soon(enter_and_push_device, stack, key, cm, results)
        yield results



import sys
from collections.abc import Callable, Awaitable
from typing import Optional, TypeVar, ParamSpec

import trio
from PySide6.QtCore import QObject, QEvent, QEventLoop
from PySide6.QtWidgets import QApplication
from outcome import Outcome


class _ReenterQtEvent(QEvent):
    def __init__(self, fn):
        super().__init__(QEvent.Type(QEvent.Type.User + 1))
        self.fn = fn


class _ReenterQtObject(QObject):
    def event(self, event):
        if event.type() == QEvent.Type.User + 1:
            assert isinstance(event, _ReenterQtEvent)
            event.fn()
            return True
        return False


_T = TypeVar("_T")
_P = ParamSpec("_P")


def run(
    async_fn: Callable[_P, Awaitable[_T]], *args: _P.args, **kwargs: _P.kwargs
) -> _T:
    """Run a Trio function in a Qt application.

    This function will run the given Trio function in the Qt event loop by spawning a
    trio loop in guest mode.
    It will block until the function completes.

    There is no need to call QApplication.exec() as this function will call it for you.

    If the function finishes successfully, the application quits and the function
    result is returned.
    If the function raises an exception, the application quits and the exception is
    raised.
    If the application is closed before the function completes, the function is
    cancelled and remaining tasks can finish before the application quits.
    """

    reenter_qt_object = _ReenterQtObject()

    outcome: Optional[Outcome] = None

    cancel_scope = trio.CancelScope()

    app = QApplication.instance()

    if app is None:
        raise RuntimeError("No QApplication instance found")

    async def wrap():
        with cancel_scope:
            await async_fn(*args, **kwargs)

    def run_sync_soon_threadsafe(fn):
        app.postEvent(reenter_qt_object, _ReenterQtEvent(fn))

    # This callback takes care of closing the QApplication when the trio function
    # completes.
    def done_callback(trio_main_outcome: Outcome):
        nonlocal outcome
        outcome = trio_main_outcome
        app.quit()

    trio.lowlevel.start_guest_run(
        wrap,
        run_sync_soon_threadsafe=run_sync_soon_threadsafe,
        done_callback=done_callback,
    )

    # We also need to handle the case where the application is closed before the
    # trio loop completes.
    def on_app_about_to_quit():
        cancel_scope.cancel()
        while outcome is None:
            app.processEvents(QEventLoop.ProcessEventsFlag.ExcludeUserInputEvents)

    app.aboutToQuit.connect(on_app_about_to_quit)

    previous_excepthook = sys.excepthook

    exception_raised = None

    def excepthook(*args):
        nonlocal exception_raised
        exception_raised = args[1]
        app.exit(-1)

    try:
        sys.excepthook = excepthook
        app.exec()
    finally:
        sys.excepthook = previous_excepthook
        app.aboutToQuit.disconnect(on_app_about_to_quit)

    if outcome is None:
        raise RuntimeError("Application exited before trio loop completed")

    try:
        if exception_raised is not None:
            raise RuntimeError(
                "An exception occurred in a Qt slot"
            ) from exception_raised
    except:
        outcome.unwrap()
        raise
    else:
        return outcome.unwrap()



"""Define devices that captures images."""

from ._compiler import CameraCompiler
from ._configuration import CameraConfiguration, CameraConfigurationType
from ._controller import CameraController
from ._proxy import CameraProxy
from ._runtime import Camera, CameraTimeoutError

__all__ = [
    "CameraConfiguration",
    "CameraConfigurationType",
    "Camera",
    "CameraTimeoutError",
    "CameraCompiler",
    "CameraProxy",
    "CameraController",
]



from __future__ import annotations

from typing import Optional, assert_never

import attrs
import pyqtgraph
from PySide6.QtWidgets import QDialog
from PySide6.QtWidgets import QWidget

from caqtus.device import DeviceName
from caqtus.session import Shot, ExperimentSessionMaker
from caqtus.types.data import DataLabel
from caqtus.types.image import ImageLabel, Image, is_image
from caqtus.utils import serialization
from .image_view_dialog_ui import Ui_ImageViewDialog
from ..single_shot_view import ShotView


@attrs.define
class ImageViewState:
    camera_name: DeviceName
    image: ImageLabel
    background: Optional[ImageLabel] = None
    colormap: Optional[str] = None
    levels: Optional[tuple[float, float]] = None


class ImageView(ShotView, pyqtgraph.ImageView):
    def __init__(
        self,
        state: ImageViewState,
        session_maker: ExperimentSessionMaker,
        parent: Optional[QWidget] = None,
    ):
        super().__init__(parent)
        self._state = state
        self.set_state(state)
        self._session_maker = session_maker
        self.getHistogramWidget().item.sigLevelChangeFinished.connect(
            self._on_levels_changed
        )

    def get_state(self) -> ImageViewState:
        return self._state

    def set_state(self, state: ImageViewState) -> None:
        if state.colormap is not None:
            colormap = pyqtgraph.colormap.get(state.colormap, source="matplotlib")
            self.setColorMap(colormap)
        if state.levels is not None:
            self.setLevels(*state.levels)

    async def display_shot(self, shot: Shot) -> None:
        camera_name = self._state.camera_name
        picture_name = self._state.image
        async with self._session_maker.async_session() as session:
            image_label = DataLabel(f"{camera_name}\\{picture_name}")
            image = await session.sequences.get_shot_data_by_label(
                shot.sequence_path, shot.index, image_label
            )
            if not is_image(image):
                raise ValueError(f"Data {image_label} is not an image.")
        self.set_image(image)

    def set_image(self, image: Image) -> None:
        match self._state.levels:
            case None:
                autoRange = True
                levels = None
                autoHistogramRange = True
            case (min_, max_):
                autoRange = False
                autoHistogramRange = False
                levels = (min_, max_)
            case _:
                assert_never(self._state.levels)
        self.setImage(
            image[::, ::-1],
            autoRange=autoRange,
            levels=levels,
            autoHistogramRange=autoHistogramRange,
        )
        self.getHistogramWidget().item.sigLevelsChanged.connect(self._on_levels_changed)

    def _on_levels_changed(self) -> None:
        if self._state.levels is not None:
            self._state.levels = self.getLevels()


class ImageViewDialog(QDialog, Ui_ImageViewDialog):
    def __init__(self):
        super().__init__()
        self.setupUi(self)


def create_image_view() -> tuple[str, ImageViewState]:
    dialog = ImageViewDialog()
    result = dialog.exec()
    if result == QDialog.DialogCode.Accepted:
        name = dialog.view_name_line_edit.text()
        state = ImageViewState(
            camera_name=DeviceName(dialog.camera_name_line_edit.text()),
            image=ImageLabel(dialog.image_line_edit.text()),
            background=(
                ImageLabel(text)
                if (text := dialog.background_line_edit.text())
                else None
            ),
        )
        return name, serialization.unstructure(state)

    else:
        raise ValueError("Dialog was cancelled.")



from __future__ import annotations

import contextlib
import datetime
import functools
import logging
import warnings
import weakref
from collections.abc import AsyncIterable, Awaitable, Callable
from typing import TypeVar

import anyio
import attrs
import tblib.pickling_support
from anyio.abc import TaskStatus
from anyio.streams.memory import MemoryObjectSendStream, MemoryObjectReceiveStream

from caqtus.device._controller import DeviceException
from caqtus.formatter import fmt
from caqtus.shot_compilation import VariableNamespace
from caqtus.types.recoverable_exceptions import ShotAttemptsExceededError
from caqtus.utils.logging import log_async_cm_decorator, log_async_cm
from .._async_utils import task_group_with_error_message
from .._shot_compiler import ShotCompilerProtocol
from .._shot_primitives import DeviceParameters, ShotData, ShotParameters
from .._shot_runner import ShotRunnerProtocol

logger = logging.getLogger(__name__)

_log_async_cm = functools.partial(log_async_cm, logger=logger)


class ShotExecutionSorter:
    """Wraps a memory object send stream to ensure that shots are executed in order."""

    def __init__(self, shot_execution_stream: MemoryObjectSendStream[DeviceParameters]):
        self._shot_execution_stream = shot_execution_stream
        self._next_shot = 0
        self._can_push_events = weakref.WeakValueDictionary[int, anyio.Event]()

    async def push(self, shot_parameters: DeviceParameters) -> None:
        """Pushes a shot to the execution queue.

        Push the shot parameters to the execution queue when the shot index matches the
        next shot to run.
        """

        shot_index = shot_parameters.index
        if shot_index != self._next_shot:
            assert shot_index > self._next_shot
            try:
                event = self._can_push_events[shot_index]
            except KeyError:
                event = anyio.Event()
                self._can_push_events[shot_index] = event
            await event.wait()

        assert shot_index == self._next_shot

        await self._shot_execution_stream.send(shot_parameters)
        self._next_shot += 1
        try:
            self._can_push_events[self._next_shot].set()
        except KeyError:
            pass


class ShotManager:
    """Manages the execution of shots.

    This object acts as an execution queue for shots on the experiment.

    When a shot is scheduled, the shot parameters are compiled into device parameters.
    The device parameters are then queued to run a shot on the experiment.
    The data produced by the shot is then yielded. The data must be consumed to allow
    further shots to be executed and scheduled.

    Args:
        shot_runner: The object that will actually execute the shots on the experiment.
        shot_compiler: The object that compiles shot parameters into device parameters.
        shot_retry_config: Specifies how to retry a shot if an error occurs.
    """

    def __init__(
        self,
        shot_runner: ShotRunnerProtocol,
        shot_compiler: ShotCompilerProtocol,
        shot_retry_config: ShotRetryConfig,
    ):
        self._shot_runner = shot_runner
        self._shot_compiler = shot_compiler
        self._shot_retry_config = shot_retry_config

        self._exit_stack = contextlib.AsyncExitStack()

    async def __aenter__(
        self,
    ) -> tuple[
        contextlib.AbstractAsyncContextManager[ShotScheduler],
        contextlib.AbstractAsyncContextManager[AsyncIterable[ShotData]],
    ]:
        """Start background tasks to compile and run shots.

        Returns:
            A tuple with two objects:
            - A context manager that allows to schedule shots.
            - A context manager that yields shot data.

            The context managers must be entered and closed before the ShotManager is
            closed because it is necessary to know when all shots have been scheduled
            and the data has been consumed.

            The data produced must be consumed to allow further shots to be executed.
        """

        await self._exit_stack.__aenter__()
        (
            shot_data_send_stream,
            shot_data_receive_stream,
        ) = anyio.create_memory_object_stream[ShotData](1)
        task_group = await self._exit_stack.enter_async_context(
            task_group_with_error_message(
                "Errors occurred while managing shots execution"
            )
        )
        (
            device_parameters_send_stream,
            device_parameters_receive_stream,
        ) = anyio.create_memory_object_stream[DeviceParameters]()
        await task_group.start(
            self.run_shots,
            self._shot_runner,
            device_parameters_receive_stream,
            shot_data_send_stream,
        )
        (
            self._shot_parameters_send_stream,
            shot_parameters_receive_stream,
        ) = anyio.create_memory_object_stream[ShotParameters]()
        await task_group.start(
            self.compile_shots,
            self._shot_compiler,
            shot_parameters_receive_stream,
            device_parameters_send_stream,
        )

        return self.scheduler(), shot_data_receive_stream

    async def __aexit__(self, exc_type, exc_value, traceback):
        return await self._exit_stack.__aexit__(exc_type, exc_value, traceback)

    async def run_shots(
        self,
        shot_runner: ShotRunnerProtocol,
        device_parameters_output_stream: MemoryObjectReceiveStream[DeviceParameters],
        shot_data_input_stream: MemoryObjectSendStream[ShotData],
        *,
        task_status: TaskStatus[None] = anyio.TASK_STATUS_IGNORED,
    ) -> None:
        # Suppress BrokenResourceError because the stream if the stream is closed due
        # to an error on the other side, we don't want to clutter the traceback.
        with contextlib.suppress(anyio.BrokenResourceError):
            async with shot_data_input_stream, device_parameters_output_stream:
                task_status.started()
                async for device_parameters in device_parameters_output_stream:
                    try:
                        shot_data = await self._run_shot_with_retry(
                            device_parameters, shot_runner
                        )
                    except Exception as e:
                        raise RuntimeError(
                            fmt(
                                "An error occurred while executing {:shot}",
                                device_parameters.index,
                            )
                        ) from e
                    await send_fast(
                        shot_data_input_stream, shot_data, "generated shot data stream"
                    )

    @log_async_cm_decorator(logger)
    @contextlib.asynccontextmanager
    async def scheduler(self):
        """Returns an object that allows to schedule shots.

        Warnings:
            It does NOT support being called several times.
        """

        async with _log_async_cm(
            self._shot_parameters_send_stream, name="shot_parameters_input_stream"
        ):
            yield ShotScheduler(self._shot_parameters_send_stream)

    async def compile_shots(
        self,
        shot_compiler: ShotCompilerProtocol,
        shot_params_receive_stream: MemoryObjectReceiveStream[ShotParameters],
        device_parameters_send_stream: MemoryObjectSendStream[DeviceParameters],
        *,
        task_status: TaskStatus[None] = anyio.TASK_STATUS_IGNORED,
    ):
        async with (
            device_parameters_send_stream,
            task_group_with_error_message(
                "Errors occurred during shot compilation"
            ) as tg,
        ):
            shot_execution_queue = ShotExecutionSorter(device_parameters_send_stream)
            async with shot_params_receive_stream:
                for _ in range(4):
                    await tg.start(
                        self._compile_shots,
                        shot_compiler,
                        shot_params_receive_stream.clone(),
                        shot_execution_queue,
                    )
            task_status.started()

    async def _compile_shots(
        self,
        shot_compiler: ShotCompilerProtocol,
        shot_params_receive_stream: MemoryObjectReceiveStream[ShotParameters],
        shot_execution_queue: ShotExecutionSorter,
        *,
        task_status: TaskStatus[None] = anyio.TASK_STATUS_IGNORED,
    ) -> None:
        # Suppress BrokenResourceError because the stream if the stream is closed due
        # to an error on the other side, we don't want to clutter the traceback.
        with contextlib.suppress(anyio.BrokenResourceError):
            async with shot_params_receive_stream:
                task_status.started()
                async for shot_params in shot_params_receive_stream:
                    result = await self._compile_shot(shot_params, shot_compiler)
                    logger.debug(
                        "Pushing shot %d to execution queue.", shot_params.index
                    )
                    await shot_execution_queue.push(result)

    async def _run_shot_with_retry(
        self, device_parameters: DeviceParameters, shot_runner: ShotRunnerProtocol
    ) -> ShotData:
        data = await _run_shot_with_retry(
            functools.partial(run_shot, device_parameters, shot_runner),
            retry_condition(self._shot_retry_config.exceptions_to_retry),
            self._shot_retry_config.number_of_attempts,
        )
        return data

    async def _compile_shot(
        self, shot_parameters: ShotParameters, shot_compiler: ShotCompilerProtocol
    ) -> DeviceParameters:
        try:
            compiled, shot_duration = await shot_compiler.compile_shot(shot_parameters)
            result = DeviceParameters(
                index=shot_parameters.index,
                shot_parameters=shot_parameters.parameters,
                device_parameters=compiled,
                timeout=shot_duration + 2,
            )
        except Exception as e:
            raise ShotCompilationError(
                fmt("An error occurred while compiling {:shot}", shot_parameters.index)
            ) from e
        return result


async def run_shot(
    device_parameters: DeviceParameters,
    shot_runner: ShotRunnerProtocol,
) -> ShotData:
    start_time = datetime.datetime.now(tz=datetime.timezone.utc)
    data = await shot_runner.run_shot(device_parameters)
    end_time = datetime.datetime.now(tz=datetime.timezone.utc)
    data = ShotData(
        index=device_parameters.index,
        start_time=start_time,
        end_time=end_time,
        variables=device_parameters.shot_parameters,
        data=data,
    )
    return data


async def _run_shot_with_retry(
    run: Callable[[], Awaitable[ShotData]],
    condition: Callable[[Exception], bool],
    number_of_attempts: int,
) -> ShotData:
    """Run a shot with retry logic.

    Args:
        run: A function to call to run the shot.
        condition: A function that determines if an exception is retriable.

            If the run function raises an exception group, the condition function will
            be tested against the exceptions in the group. If the condition function
            returns True for all the sub-exceptions, the shot will be retried.

        number_of_attempts: The maximum number of times to retry the shot if an error
            occurs.

            Must be >= 1.

    Returns:
        The data produced by the shot.

    Raises:
        ShotAttemptsExceededError: If the shot could not be executed after the number of
            attempts specified.
    """

    if number_of_attempts < 1:
        raise ValueError("number_of_attempts must be >= 1")

    errors: list[ExceptionGroup] = []

    for attempt in range(number_of_attempts):
        try:
            return await run()
        except BaseExceptionGroup as exc_group:
            retriable, others = exc_group.split(condition)
            if others:
                raise
            errors.append(
                ExceptionGroup(
                    f"Attempt {attempt+1}/{number_of_attempts} failed",
                    retriable.exceptions,
                )
            )
            logger.warning(
                f"Attempt {attempt+1}/{number_of_attempts} failed", exc_info=exc_group
            )
    raise ShotAttemptsExceededError(
        f"Could not execute shot after {number_of_attempts} attempts", errors
    )


def retry_condition(
    retriable_exceptions: tuple[type[Exception], ...]
) -> Callable[[Exception], bool]:
    def _retry_condition(e: Exception) -> bool:
        return isinstance(e, DeviceException) and isinstance(
            e.__cause__, retriable_exceptions
        )

    return _retry_condition


@tblib.pickling_support.install
class ShotCompilationError(RuntimeError):
    """Error raised when an error occurs while compiling a shot."""

    pass


@attrs.define
class ShotRetryConfig:
    """Specifies how to retry a shot if an error occurs.

    Attributes:
        exceptions_to_retry: If an exception occurs while running a shot, it will be
        retried if it is an instance of one of the exceptions in this tuple.
        number_of_attempts: The number of times to retry a shot if an error occurs.
    """

    exceptions_to_retry: tuple[type[Exception], ...] = attrs.field(
        factory=tuple,
        eq=False,
        on_setattr=attrs.setters.validate,
    )
    number_of_attempts: int = attrs.field(default=1, eq=False)


class ShotScheduler:
    def __init__(
        self,
        shot_parameters_input_stream: MemoryObjectSendStream[ShotParameters],
    ):
        self._shot_parameters_input_stream = shot_parameters_input_stream
        self._current_shot = 0

    async def schedule_shot(self, shot_variables: VariableNamespace) -> None:
        shot_parameters = ShotParameters(
            index=self._current_shot, parameters=shot_variables
        )
        with contextlib.suppress(anyio.BrokenResourceError):
            await self._shot_parameters_input_stream.send(shot_parameters)
        self._current_shot += 1


_T = TypeVar("_T")


async def send_fast(
    stream: MemoryObjectSendStream[_T], value: _T, stream_name: str
) -> None:
    try:
        return stream.send_nowait(value)
    except anyio.WouldBlock:
        try:
            with anyio.fail_after(1):
                await stream.send(value)
        except TimeoutError:
            raise RuntimeError(f"Could not push value to {stream_name}") from None
        message = (
            f"Could not push value to {stream_name} instantly. "
            "The data consumer can't keep up, which might cause a slowdown of the "
            "experiment."
        )
        warnings.warn(message, stacklevel=2)
        logger.warning(message)



import abc
import contextlib
from collections.abc import Callable, Iterator, AsyncIterator
from typing import Protocol, LiteralString

from ._proxy import Proxy


class AsyncConverter(Protocol):
    """Transform synchronous calls into asynchronous ones.

    To support subprocess calls, all function arguments must be either pickleable or a
    proxy to a remote object.
    """

    @abc.abstractmethod
    async def call[
        T,
    ](self, fun: Callable[..., T], *args, **kwargs) -> T:
        """Call a function asynchronously.

        Args:
            fun: The function to call.
            args: The positional arguments to pass to the function.
            kwargs: The keyword arguments to pass to the function.

        Returns:
            A copy of the return value of the function.

        Warning:
            To support subprocess calls, all arguments must be
            and the return value must be
            pickleable.
        """

        raise NotImplementedError

    @abc.abstractmethod
    async def call_method(
        self,
        obj,
        method: LiteralString,
        *args,
        **kwargs,
    ):
        """Call an object's method asynchronously.

        Args:
            obj: The object on which to call the method.

                It must be either pickleable or a proxy to a remote object.

            method: The name of the method to call.
            args: The positional arguments to pass to the method.

                They must be pickleable.

            kwargs: The keyword arguments to pass to the method.

                They must be pickleable.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def call_method_proxy_result(
        self,
        obj,
        method: LiteralString,
        *args,
        **kwargs,
    ) -> contextlib.AbstractAsyncContextManager[Proxy]:
        """Call an object's method and return a proxy to the result.

        Args:
            obj: The object on which to call the method.

                Must be pickleable or a proxy to a remote object.

            method: The name of the method to call.
            args: The positional arguments to pass to the method.

                Must be pickleable.

            kwargs: The keyword arguments to pass to the method.

                Must be pickleable.

        Returns:
            An async context manager that yields a proxy to the result of the method
            call.
        """

        raise NotImplementedError

    @abc.abstractmethod
    async def get_attribute(self, obj, attribute: LiteralString):
        """Get an object's attribute asynchronously.

        Args:
            obj: The object from which to get the attribute.

                Must be pickleable or a proxy to a remote object.

            attribute: The name of the attribute to get.

        Returns:
            A copy of the attribute's value.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def call_proxy_result[
        T
    ](
        self, fun: Callable[..., T], *args, **kwargs
    ) -> contextlib.AbstractAsyncContextManager[Proxy[T]]:
        """Call a function and return a proxy to the result."""

        raise NotImplementedError

    @abc.abstractmethod
    def async_context_manager[
        T
    ](
        self, cm_proxy: Proxy[contextlib.AbstractContextManager[T]]
    ) -> contextlib.AbstractAsyncContextManager[Proxy[T]]:
        """Wrap an async context manager proxy.

        Returns:
            An async context manager that yields a proxy to the result of context
            manager's __enter__ method.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def async_iterator[T](self, proxy: Proxy[Iterator[T]]) -> AsyncIterator[T]:
        """Iterate over a remote iterator asynchronously."""

        raise NotImplementedError



import datetime
from typing import Protocol, TypeVar, Optional

from caqtus.session import Sequence, PureSequencePath

T = TypeVar("T", covariant=True)


class SequenceFunction(Protocol[T]):
    def __call__(self, sequence: Sequence) -> T: ...


# noinspection PyPep8Naming
class cache_per_sequence:
    def __init__(self, func: SequenceFunction[T]) -> None:
        self.func = func
        self.cache: dict[PureSequencePath, tuple[Optional[datetime.datetime], T]] = {}

    def __call__(self, sequence: Sequence) -> T:
        path = sequence.path
        last_start_time = sequence.get_start_time()
        if path not in self.cache:
            result = self.func(sequence)
            self.cache[path] = (last_start_time, result)
            return result
        else:
            cached_start_time, result = self.cache[path]
            sequence_has_been_started = last_start_time is not None
            sequence_has_been_reset = last_start_time != cached_start_time
            if sequence_has_been_reset or not sequence_has_been_started:
                result = self.func(sequence)
                self.cache[path] = (last_start_time, result)
                return result
            else:
                return result



"""Allows to interact with the storage of the experiment."""

from . import sql
from ._device_configuration_collection import DeviceConfigurationCollection
from ._exception_summary import TracebackSummary
from ._experiment_session import ExperimentSession
from ._path import PureSequencePath, InvalidPathFormatError
from ._path_hierarchy import (
    PathError,
    PathNotFoundError,
    PathIsRootError,
    PathHasChildrenError,
)
from ._path_hierarchy import PathHierarchy
from ._sequence import Sequence, Shot
from ._sequence_collection import (
    SequenceCollection,
    PathIsSequenceError,
    PathIsNotSequenceError,
    DataNotFoundError,
    SequenceStateError,
    InvalidStateTransitionError,
    SequenceNotEditableError,
    ShotNotFoundError,
    SequenceNotCrashedError,
)
from ._session_maker import ExperimentSessionMaker
from ._state import State
from .async_session import AsyncExperimentSession

__all__ = [
    "ExperimentSession",
    "Sequence",
    "State",
    "Shot",
    "ExperimentSessionMaker",
    "PathHierarchy",
    "PureSequencePath",
    "InvalidPathFormatError",
    "SequenceCollection",
    "DeviceConfigurationCollection",
    "AsyncExperimentSession",
    "PathIsSequenceError",
    "PathIsNotSequenceError",
    "DataNotFoundError",
    "SequenceStateError",
    "InvalidStateTransitionError",
    "SequenceNotEditableError",
    "ShotNotFoundError",
    "PathError",
    "PathNotFoundError",
    "PathIsRootError",
    "PathHasChildrenError",
    "TracebackSummary",
    "SequenceNotCrashedError",
    "sql",
]



"""Contains the GUI components of the application.

It is implemented as a separate package build on top of the core functionalities.
"""

from . import condetrol, snapshot

__all__ = ["condetrol", "snapshot"]



import decimal
from typing import TypeVar, Generic, Optional

from PySide6.QtWidgets import QWidget, QVBoxLayout, QSpinBox

from caqtus.device.sequencer import SequencerConfiguration, TimeStep
from ._trigger_selector import TriggerSelector
from .channels_widget import SequencerChannelWidget
from .._device_configuration_editor import FormDeviceConfigurationEditor

S = TypeVar("S", bound=SequencerConfiguration)


class SequencerConfigurationEditor(FormDeviceConfigurationEditor[S], Generic[S]):
    """A widget to edit a sequencer configuration.

    Args:
        device_configuration: The initial configuration to display.
        time_step_increment: The smallest value (in ns) by which it is possible to
            change the time step in the editor.
            Only time step that are multiple of this value can be selected0
        smallest_increment_multiple: The smallest multiple of the increment allowed.
        largest_increment_multiple: The largest multiple of the increment allowed.
        parent: The parent widget.

    Attributes:
        time_step_widget: A widget to select the time step of the sequencer.
        trigger_selector: A widget to select the trigger of the sequencer.
        channels_widget: A widget to edit the channels of the sequencer.
    """

    def __init__(
        self,
        device_configuration: S,
        time_step_increment: TimeStep,
        smallest_increment_multiple: int = 1,
        largest_increment_multiple: int = 100000,
        parent: Optional[QWidget] = None,
    ):
        super().__init__(device_configuration, parent)

        self.time_step_widget = TimeStepEditor(
            time_step_increment, smallest_increment_multiple, largest_increment_multiple
        )
        self.time_step_widget.set_time_step(device_configuration.time_step)
        self.form.addRow("Time step", self.time_step_widget)

        self.trigger_selector = TriggerSelector(parent=self)
        self.trigger_selector.set_trigger(self.device_configuration.trigger)
        self.form.addRow("Trigger", self.trigger_selector)

        self.channels_widget = SequencerChannelWidget(
            self.device_configuration.channels, self
        )
        self.form.addRow("Channels", self.channels_widget)

    def get_configuration(self) -> S:
        config = super().get_configuration()
        config.time_step = self.time_step_widget.read_time_step()
        config.channels = self.channels_widget.get_channel_configurations()
        config.trigger = self.trigger_selector.get_trigger()
        return config


class TimeStepEditor(QWidget):
    def __init__(
        self,
        increment: TimeStep,
        smallest_multiple: int = 1,
        largest_multiple: int = 100000,
        parent: Optional[QWidget] = None,
    ):
        super().__init__(parent)
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        self.increment = increment
        self.smallest_multiple = smallest_multiple
        self.largest_multiple = largest_multiple

        self.spin_box = QSpinBox()
        self.spin_box.setRange(smallest_multiple, largest_multiple)
        self.spin_box.setSuffix(format_suffix(increment))
        self.spin_box.setStepType(QSpinBox.StepType.AdaptiveDecimalStepType)
        layout.addWidget(self.spin_box)

    def set_time_step(self, time_step: TimeStep):
        div, mod = divmod(time_step, self.increment)
        if mod != 0:
            raise ValueError(
                f"Time step {time_step} is not a multiple of the increment "
                f"{self.increment}"
            )
        if not (self.smallest_multiple <= div <= self.largest_multiple):
            raise ValueError(
                f"Time step {time_step} is not between {self.smallest_multiple} and "
                f"{self.largest_multiple}"
            )
        numerator, denominator = div.as_integer_ratio()
        assert denominator == 1
        self.spin_box.setValue(numerator)

    def read_time_step(self) -> TimeStep:
        multiple = self.spin_box.value()
        return self.increment * multiple


def format_suffix(time_step: TimeStep) -> str:
    if time_step < 1:
        return f" × {time_step} ns"
    elif time_step == 1:
        return " ns"
    elif 1 < time_step < 1000:
        return f" × {time_step} ns"
    elif 1e3 <= time_step < 1e6:
        return f" × {time_step / decimal.Decimal('1e3')} µs"
    elif 1e6 <= time_step < 1e9:
        return f" × {time_step / decimal.Decimal('1e6')} ms"
    elif 1e9 <= time_step:
        return f" × {time_step / decimal.Decimal('1e9')} s"



from typing import assert_never

import attrs

from caqtus.types.expression import Expression
from caqtus.utils import serialization
from .timelane import TimeLane


class Ramp:
    def __eq__(self, other):
        return isinstance(other, Ramp)

    def __repr__(self):
        return "Ramp()"


def unstructure_union(union: Expression | Ramp):
    if isinstance(union, Expression):
        return {"expression": union.body}
    elif isinstance(union, Ramp):
        return "ramp"
    else:
        assert_never(union)


def structure_union(data, _) -> Expression | Ramp:
    if isinstance(data, dict):
        return Expression(data["expression"])
    elif data == "ramp":
        return Ramp()
    else:
        raise ValueError(f"Invalid union value: {data}")


serialization.register_structure_hook(Expression | Ramp, structure_union)
serialization.register_unstructure_hook(Expression | Ramp, unstructure_union)


@attrs.define(init=False, eq=False, repr=False)
class AnalogTimeLane(TimeLane[Expression | Ramp]):
    """A time lane that represents analog values changing over time.

    The values of an analog time lane can be a placeholder expression or a ramp.
    The placeholder expression represents a values that is constant for the given steps.
    The ramp represents a linear change from the previous step to the next step.

    The conditions below are not checked by the class, but code that evaluates the lane
    should enforce them:

    #. A ramp should not be the first or last value of the lane.
    #. There should not be two consecutive ramps in the lane. Use a single ramp block
        spanning multiple steps instead.
    #. Expressions representing values with units should all have the same dimension
        for a given lane.

    Examples:

    .. code-block:: python

        from caqtus.session.shot import AnalogTimeLane
        from caqtus.types.expression import Expression

        # Creates an analog time lane with known values
        lane = AnalogTimeLane([Expression("0 MHz")] + [Ramp()] * 2 + [Expression("10 MHz")])

        # Creates an analog time lane with a placeholder expression
        lane = AnalogTimeLane([Expression("2 * x"), Expression("y")])
    """

    pass


def unstructure_hook(lane: AnalogTimeLane):
    return {
        "spanned_values": serialization.unstructure(
            lane._spanned_values, list[tuple[Expression | Ramp, int]]
        )
    }


def structure_hook(data, _) -> AnalogTimeLane:
    structured = serialization.structure(
        data["spanned_values"], list[tuple[Expression | Ramp, int]]
    )
    return AnalogTimeLane.from_spanned_values(structured)


serialization.register_structure_hook(AnalogTimeLane, structure_hook)
serialization.register_unstructure_hook(AnalogTimeLane, unstructure_hook)



from collections.abc import Mapping

from ._units import UNITS, ureg, Unit
from ..variable_name import VariableName

units: Mapping[VariableName, Unit] = {
    VariableName(unit): getattr(ureg, unit) for unit in UNITS
}



from __future__ import annotations

import abc
import contextlib
from collections.abc import Callable, AsyncGenerator
from collections.abc import Mapping
from typing import Protocol

from caqtus.device import DeviceController
from caqtus.types.data import DataLabel, Data
from ._initialize_devices import create_devices
from ._shot_compiler import ShotCompilerProtocol
from ._shot_event_dispatcher import DeviceRunConfig, ShotEventDispatcher
from ._shot_primitives import DeviceParameters
from .device_manager_extension import DeviceManagerExtensionProtocol
from ..device import DeviceName, DeviceConfiguration
from ..device.remote import DeviceProxy
from ..shot_compilation import SequenceContext


class ShotRunnerProtocol(Protocol):
    """Interface for running a shot."""

    @abc.abstractmethod
    async def run_shot(
        self, shot_parameters: DeviceParameters
    ) -> Mapping[DataLabel, Data]:
        """Run a shot.

        Args:
            shot_parameters: The parameters for the shot.
        """

        ...


type ShotRunnerFactory = Callable[
    [SequenceContext, ShotCompilerProtocol, DeviceManagerExtensionProtocol],
    contextlib.AbstractAsyncContextManager[ShotRunnerProtocol, None],
]
"""A function called to create a object that can run a shot."""


class ShotRunner(ShotRunnerProtocol):
    def __init__(
        self,
        devices: Mapping[DeviceName, DeviceProxy],
        controller_types: Mapping[DeviceName, type[DeviceController]],
    ):
        if set(devices.keys()) != set(controller_types.keys()):
            raise ValueError("The devices and controller_types must have the same keys")
        self.devices = devices
        self.controller_types = controller_types

    async def run_shot(
        self,
        shot_params: DeviceParameters,
    ) -> Mapping[DataLabel, Data]:
        event_dispatcher = ShotEventDispatcher(
            {
                name: DeviceRunConfig(
                    device=self.devices[name],
                    controller_type=self.controller_types[name],
                    parameters=shot_params.device_parameters[name],
                )
                for name in self.devices
            }
        )
        return await event_dispatcher.run_shot(shot_params.timeout)


@contextlib.asynccontextmanager
async def create_shot_runner(
    sequence_context: SequenceContext,
    shot_compiler: ShotCompilerProtocol,
    device_manager_extension: DeviceManagerExtensionProtocol,
) -> AsyncGenerator[ShotRunner, None]:
    """Creates and acquires resources for running a shot.

    Returns:
        A context manager that yields a shot runner.
    """

    initialization_parameters = shot_compiler.compile_initialization_parameters()

    device_configurations_in_use = {
        name: sequence_context.get_device_configuration(name)
        for name in initialization_parameters
    }

    device_types = {
        name: device_manager_extension.get_device_type(config)
        for name, config in device_configurations_in_use.items()
    }

    async with create_devices(
        initialization_parameters=initialization_parameters,
        device_configs=device_configurations_in_use,
        device_types=device_types,
        device_manager_extension=device_manager_extension,
    ) as devices_in_use:
        shot_runner = _create_shot_runner(
            device_proxies=devices_in_use,
            device_configurations=device_configurations_in_use,
            device_manager_extension=device_manager_extension,
        )

        yield shot_runner


def _create_shot_runner(
    device_proxies: Mapping[DeviceName, DeviceProxy],
    device_configurations: Mapping[DeviceName, DeviceConfiguration],
    device_manager_extension: DeviceManagerExtensionProtocol,
) -> ShotRunner:
    device_controller_types = {
        name: device_manager_extension.get_device_controller_type(
            device_configurations[name]
        )
        for name in device_proxies
    }
    shot_runner = ShotRunner(device_proxies, device_controller_types)
    return shot_runner



import attrs


@attrs.define
class InsecureRPCConfiguration:
    host: str = attrs.field(converter=str, on_setattr=attrs.setters.convert)
    port: int = attrs.field(converter=int, on_setattr=attrs.setters.convert)


RPCConfiguration = InsecureRPCConfiguration



import abc
from typing import Protocol

from .async_session import AsyncExperimentSession
from ._experiment_session import ExperimentSession


class ExperimentSessionMaker(Protocol):
    """Used to create a new experiment sessions."""

    @abc.abstractmethod
    def __call__(self) -> ExperimentSession:
        """Create a new experiment session."""

        raise NotImplementedError

    def session(self) -> ExperimentSession:
        """Create a new experiment session."""

        return self()

    @abc.abstractmethod
    def async_session(self) -> AsyncExperimentSession:
        """Create a new asynchronous experiment session."""

        raise NotImplementedError



"""Define classes and functions to evaluate output for a sequencer."""

from ._compiler import (
    SequencerCompiler,
    InstructionCompilationParameters,
    compile_parallel_instructions,
)
from ..channel_commands._channel_sources._trigger_compiler import (
    TriggerableDeviceCompiler,
)

__all__ = [
    "SequencerCompiler",
    "TriggerableDeviceCompiler",
    "InstructionCompilationParameters",
    "compile_parallel_instructions",
]



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'calibrated_analog_mapping_widget.ui'
##
## Created by: Qt User Interface Compiler version 6.6.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QFormLayout,
    QHBoxLayout,
    QHeaderView,
    QLabel,
    QLineEdit,
    QSizePolicy,
    QSpacerItem,
    QTabWidget,
    QTableView,
    QToolButton,
    QVBoxLayout,
    QWidget,
)


class Ui_CalibratedAnalogMappingWigdet(object):
    def setupUi(self, CalibratedAnalogMappingWigdet):
        if not CalibratedAnalogMappingWigdet.objectName():
            CalibratedAnalogMappingWigdet.setObjectName("CalibratedAnalogMappingWigdet")
        CalibratedAnalogMappingWigdet.resize(758, 466)
        self.horizontalLayout_2 = QHBoxLayout(CalibratedAnalogMappingWigdet)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.tabWidget = QTabWidget(CalibratedAnalogMappingWigdet)
        self.tabWidget.setObjectName("tabWidget")
        self.tab_2 = QWidget()
        self.tab_2.setObjectName("tab_2")
        self.verticalLayout_3 = QVBoxLayout(self.tab_2)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.verticalLayout = QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.formLayout = QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.inputUnitLabel = QLabel(self.tab_2)
        self.inputUnitLabel.setObjectName("inputUnitLabel")

        self.formLayout.setWidget(0, QFormLayout.LabelRole, self.inputUnitLabel)

        self.inputUnitLineEdit = QLineEdit(self.tab_2)
        self.inputUnitLineEdit.setObjectName("inputUnitLineEdit")

        self.formLayout.setWidget(0, QFormLayout.FieldRole, self.inputUnitLineEdit)

        self.outputUnitLabel = QLabel(self.tab_2)
        self.outputUnitLabel.setObjectName("outputUnitLabel")

        self.formLayout.setWidget(1, QFormLayout.LabelRole, self.outputUnitLabel)

        self.outputUnitLineEdit = QLineEdit(self.tab_2)
        self.outputUnitLineEdit.setObjectName("outputUnitLineEdit")

        self.formLayout.setWidget(1, QFormLayout.FieldRole, self.outputUnitLineEdit)

        self.verticalLayout.addLayout(self.formLayout)

        self.tableView = QTableView(self.tab_2)
        self.tableView.setObjectName("tableView")
        self.tableView.horizontalHeader().setStretchLastSection(True)
        self.tableView.verticalHeader().setVisible(False)

        self.verticalLayout.addWidget(self.tableView)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.add_button = QToolButton(self.tab_2)
        self.add_button.setObjectName("add_button")

        self.horizontalLayout.addWidget(self.add_button)

        self.remove_button = QToolButton(self.tab_2)
        self.remove_button.setObjectName("remove_button")

        self.horizontalLayout.addWidget(self.remove_button)

        self.horizontalSpacer = QSpacerItem(
            40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum
        )

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.verticalLayout.addLayout(self.horizontalLayout)

        self.verticalLayout_3.addLayout(self.verticalLayout)

        self.tabWidget.addTab(self.tab_2, "")

        self.horizontalLayout_2.addWidget(self.tabWidget)

        self.retranslateUi(CalibratedAnalogMappingWigdet)

        self.tabWidget.setCurrentIndex(0)

        QMetaObject.connectSlotsByName(CalibratedAnalogMappingWigdet)

    # setupUi

    def retranslateUi(self, CalibratedAnalogMappingWigdet):
        CalibratedAnalogMappingWigdet.setWindowTitle(
            QCoreApplication.translate("CalibratedAnalogMappingWigdet", "Form", None)
        )
        self.inputUnitLabel.setText(
            QCoreApplication.translate(
                "CalibratedAnalogMappingWigdet", "Input unit", None
            )
        )
        self.inputUnitLineEdit.setPlaceholderText(
            QCoreApplication.translate("CalibratedAnalogMappingWigdet", "None", None)
        )
        self.outputUnitLabel.setText(
            QCoreApplication.translate(
                "CalibratedAnalogMappingWigdet", "Output unit", None
            )
        )
        self.outputUnitLineEdit.setPlaceholderText(
            QCoreApplication.translate("CalibratedAnalogMappingWigdet", "None", None)
        )
        self.add_button.setText(
            QCoreApplication.translate("CalibratedAnalogMappingWigdet", "...", None)
        )
        self.remove_button.setText(
            QCoreApplication.translate("CalibratedAnalogMappingWigdet", "...", None)
        )
        self.tabWidget.setTabText(
            self.tabWidget.indexOf(self.tab_2),
            QCoreApplication.translate("CalibratedAnalogMappingWigdet", "Values", None),
        )

    # retranslateUi



from typing import Optional

from PySide6 import QtWidgets

from caqtus.session import TracebackSummary
from ._exception_tree import create_exception_tree
from .exception_dialog_ui import Ui_ExceptionDialog
from ...qtutil import HTMLItemDelegate


class ExceptionDialog(QtWidgets.QDialog, Ui_ExceptionDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setupUi(self)
        self.exception_tree.setItemDelegateForColumn(2, HTMLItemDelegate(self))
        self.exception_tree.setColumnCount(3)

    def set_exception(self, tb_summary: Optional[TracebackSummary]):
        self.exception_tree.clear()
        if tb_summary:
            tree = create_exception_tree(tb_summary)
            self.exception_tree.addTopLevelItems(tree)
        self.exception_tree.expandAll()
        for column in range(self.exception_tree.columnCount()):
            self.exception_tree.resizeColumnToContents(column)
        self.exception_tree.hideColumn(1)
        self.setWindowTitle("Error")

    def set_message(self, message: str):
        self.exception_label.setText(message)



import ast
import re
from collections.abc import Mapping
from functools import cached_property
from typing import Optional, Any

import numpy
import token_utils

from caqtus.utils import serialization
from ..recoverable_exceptions import EvaluationError
from ..units import units
from ..variable_name import DottedVariableName, VariableName

EXPRESSION_REGEX = re.compile(".*")


def square_wave(t, period, duty_cycle=0.5, low=0, high=1):
    x = t / period
    x = x - numpy.floor(x)
    is_high = x < duty_cycle
    result = numpy.where(is_high, high, low)
    return result


#: Built-in functions and constants available in expressions.
BUILTINS: Mapping[str, Any] = {
    "abs": numpy.abs,
    "arccos": numpy.arccos,
    "arcsin": numpy.arcsin,
    "arctan": numpy.arctan,
    "arctan2": numpy.arctan2,
    "ceil": numpy.ceil,
    "cos": numpy.cos,
    "cosh": numpy.cosh,
    "degrees": numpy.degrees,
    "e": numpy.e,
    "exp": numpy.exp,
    "floor": numpy.floor,
    "log": numpy.log,
    "log10": numpy.log10,
    "log2": numpy.log2,
    "pi": numpy.pi,
    "radians": numpy.radians,
    "sin": numpy.sin,
    "sinh": numpy.sinh,
    "sqrt": numpy.sqrt,
    "tan": numpy.tan,
    "tanh": numpy.tanh,
    "square_wave": square_wave,
    "max": max,
    "min": min,
    "Enabled": True,
    "Disabled": False,
} | {str(name): value for name, value in units.items()}


class Expression:
    """Represents a mathematical expression that can be evaluated in a given context.

    This class is a wrapper around a python string expression that can be evaluated
    later when the values of the variables it depends on are known.

    The expression is immutable.

    If the expression contains syntax errors, they only will be raised when the
    expression is evaluated.

    The expression must be a valid python expression, with some exceptions:

    * The % symbol is understood as a multiplication by 0.01.
    * The ° symbol is understood as the degree symbol and will be replaced by the
      name `deg` in the expression.
    * Implicit multiplication is allowed, so that "a b" will be parsed as "a * b".
    """

    def __init__(self, body: str):
        if not isinstance(body, str):
            raise TypeError(f"Expression body must be a string, got {type(body)}")
        self._body = body

    @property
    def body(self) -> str:
        """The string representation of the expression."""

        return self._body

    def __repr__(self) -> str:
        return f"Expression('{self.body}')"

    def __str__(self) -> str:
        return self.body

    def evaluate(self, variables: Mapping[DottedVariableName, Any]) -> Any:
        """Evaluate an expression on specific values for its variables.

        Args:
            variables: The context in which the expression will be evaluated.
                The variables the expression depends upon will be replaced by the values
                in this mapping.
                The mapping is also concatenated with some built-in functions and
                constants.

        Returns:
            The result of the evaluation. The type of the result depends on the body of
            the expression and the values of the upstream variables.

        Raises:
            EvaluationError: if an error occurred during evaluation.
        """

        return self._evaluate({str(expr): variables[expr] for expr in variables})

    @cached_property
    def upstream_variables(self) -> frozenset[VariableName]:
        """Return the name of the other variables the expression depend on."""

        variables = set()

        class FindNameVisitor(ast.NodeVisitor):
            def visit_Name(self, node: ast.Name):
                if isinstance(node.ctx, ast.Load):
                    if node.id not in BUILTINS:
                        variables.add(VariableName(node.id))

        FindNameVisitor().visit(self._ast)
        return frozenset(variables)

    def check_syntax(self) -> Optional[SyntaxError]:
        """Force parsing of the expression.

        It is not necessary to call this method explicitly, as the expression will be
        parsed automatically when it is evaluated. However, this method can be used
        to force the parsing to happen at a specific time, for example to catch
        syntax errors early.

        Returns:
            None if the expression is valid, or a SyntaxError otherwise.
        """

        try:
            self._parse_ast()
        except SyntaxError as error:
            return error

    def _evaluate(self, variables: Mapping[str, Any]) -> Any:
        try:
            value = eval(self._code, {"__builtins__": BUILTINS}, variables)
        except Exception as error:
            raise EvaluationError(f"Could not evaluate <{self.body}>") from error
        return value

    @cached_property
    def _ast(self) -> ast.Expression:
        """Computes the abstract syntax tree for this expression"""

        return self._parse_ast()

    def _parse_ast(self) -> ast.Expression:
        expr = self.body

        expr = expr.replace("%", "*(1e-2)")
        expr = expr.replace("°", "*deg")
        expr = add_implicit_multiplication(expr)

        return ast.parse(expr, mode="eval")

    @cached_property
    def _code(self):
        return compile(self._ast, filename="<string>", mode="eval")

    def __eq__(self, other):
        if isinstance(other, Expression):
            return self.body == other.body
        else:
            return NotImplemented

    def __getstate__(self):
        return {"body": self.body}

    def __setstate__(self, state):
        self.__init__(**state)


serialization.register_unstructure_hook(Expression, lambda expr: expr.body)
serialization.register_structure_hook(Expression, lambda body, _: Expression(body))


def add_implicit_multiplication(source: str) -> str:
    """This adds a multiplication symbol where it would be understood as
    being implicit by the normal way algebraic equations are written but would
    be a SyntaxError in Python. Thus we have::

        2n -> 2*n
        n 2 -> n* 2
        2(a+b) -> 2*(a+b)
        (a+b)2 -> (a+b)*2
        2 3 -> 2* 3
        m n -> m* n
        (a+b)c -> (a+b)*c

    The obvious one (in algebra) being left out is something like ``n(...)``
    which is a function call - and thus valid Python syntax.
    """

    tokens = token_utils.tokenize(source)
    if not tokens:
        return tokens

    prev_token = tokens[0]
    new_tokens = [prev_token]

    for token in tokens[1:]:
        if (
            (
                prev_token.is_number()
                and (token.is_identifier() or token.is_number() or token == "(")
            )
            or (
                prev_token.is_identifier()
                and (token.is_identifier() or token.is_number())
            )
            or (prev_token == ")" and (token.is_identifier() or token.is_number()))
        ):
            new_tokens.append("*")
        new_tokens.append(token)
        prev_token = token

    return token_utils.untokenize(new_tokens)



import functools
from collections.abc import Set, Iterable
from typing import Optional

import yaml
from PySide6 import QtCore, QtWidgets
from PySide6.QtCore import Qt, QPersistentModelIndex, QModelIndex
from PySide6.QtGui import QKeySequence, QShortcut, QAction, QFont, QGuiApplication
from PySide6.QtWidgets import QWidget, QTreeView, QAbstractItemView, QMenu

from caqtus.gui.qtutil import block_signals
from caqtus.types.expression import Expression
from caqtus.types.iteration import (
    StepsConfiguration,
    VariableDeclaration,
    ExecuteShot,
    LinspaceLoop,
    ArangeLoop,
)
from caqtus.types.variable_name import DottedVariableName
from caqtus.utils import serialization
from .delegate import StepDelegate
from .steps_model import StepsModel
from ..sequence_iteration_editor import SequenceIterationEditor
from ..._icons import get_icon


def create_variable_declaration():
    return VariableDeclaration(
        variable=DottedVariableName("new_variable"), value=Expression("...")
    )


def create_linspace_loop():
    return LinspaceLoop(
        variable=DottedVariableName("new_variable"),
        start=Expression("..."),
        stop=Expression("..."),
        num=10,
        sub_steps=[],
    )


def create_arange_loop():
    return ArangeLoop(
        variable=DottedVariableName("new_variable"),
        start=Expression("..."),
        stop=Expression("..."),
        step=Expression("..."),
        sub_steps=[],
    )


class StepsIterationEditor(QTreeView, SequenceIterationEditor[StepsConfiguration]):
    """Editor for the steps of a sequence.

    .. figure:: screenshot_StepsIterationEditor.png
       :align: center

       Example of a StepsIterationEditor.

    Args:
        parent: The parent widget of the editor.

    Signals:
        iteration_edited: Emitted when the iteration displayed by the editor is changed
            by the user.
            This signal is not emitted when the iteration is set programmatically.
    """

    iteration_edited = QtCore.Signal(StepsConfiguration)

    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self._model = StepsModel(StepsConfiguration([]))
        self.setModel(self._model)
        self.expandAll()
        self.header().hide()
        self._delegate = StepDelegate(self)
        self.setItemDelegate(self._delegate)

        self.setDragEnabled(True)
        self.setAcceptDrops(True)
        self.setDropIndicatorShown(True)
        self.setDragDropMode(QAbstractItemView.DragDropMode.DragDrop)
        self.setDefaultDropAction(Qt.DropAction.MoveAction)

        self.delete_shortcut = QShortcut(QKeySequence("Delete"), self)
        self.delete_shortcut.activated.connect(self.delete_selected)

        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.show_context_menu)

        self._model.dataChanged.connect(self._emit_iteration_edited)
        self._model.rowsInserted.connect(self._emit_iteration_edited)
        self._model.rowsRemoved.connect(self._emit_iteration_edited)
        self._model.modelReset.connect(self._emit_iteration_edited)

        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)

        self.select_all_shortcut = QShortcut(QKeySequence("Ctrl+A"), self)
        self.select_all_shortcut.activated.connect(self.selectAll)
        self.copy_shortcut = QShortcut(QKeySequence("Ctrl+C"), self)
        self.copy_shortcut.activated.connect(self.copy_selection)
        self.cut_shortcut = QShortcut(QKeySequence("Ctrl+X"), self)
        self.cut_shortcut.activated.connect(self.cut_selected)
        self.paste_shortcut = QShortcut(QKeySequence("Ctrl+V"), self)
        self.paste_shortcut.activated.connect(self.paste)

        font = QFont("JetBrains Mono")
        font.setPixelSize(13)
        self.setFont(font)

        self.toolbar = QtWidgets.QToolBar()
        self.copy_to_clipboard_action = self.toolbar.addAction(
            get_icon("copy", self.palette().buttonText().color()), "Copy to clipboard"
        )
        self.copy_to_clipboard_action.triggered.connect(self.copy_to_clipboard)
        self.paste_from_clipboard_action = self.toolbar.addAction(
            get_icon("paste", self.palette().buttonText().color()),
            "Paste from clipboard",
        )
        self.paste_from_clipboard_action.triggered.connect(self.paste_from_clipboard)

    def _emit_iteration_edited(self, *args, **kwargs):
        self.iteration_edited.emit(self.get_iteration())
        self.expandAll()

    def get_iteration(self) -> StepsConfiguration:
        """Returns the iteration currently displayed by the editor.

        This method returns a copy of the iteration, so the caller can modify it
        without affecting the editor.
        """

        return self._model.get_steps()

    def set_iteration(self, iteration: StepsConfiguration):
        """Updates the iteration displayed by the editor.

        This method does not emit the signal iteration_edited.
        """

        with block_signals(self):
            self._model.set_steps(iteration)
        self.expandAll()

    def set_available_parameter_names(self, parameter_names: Set[DottedVariableName]):
        """Sets the names that can be used in the iteration.

        The names are used to populate the auto-completion of variable names when the
        user try to edit the name of a variable or a loop.
        """

        self._delegate.set_available_names(parameter_names)

    def set_read_only(self, read_only: bool) -> None:
        """Sets the editor in read-only mode.

        When the editor is in read-only mode, the user cannot edit the steps.

        Even if the editor is in read-only mode, the iteration can still be set
        programmatically with :meth:`set_iteration`.
        """

        self._model.set_read_only(read_only)
        self.paste_from_clipboard_action.setEnabled(not read_only)

    def is_read_only(self) -> bool:
        """Returns whether the editor is in read-only mode."""

        return self._model.is_read_only()

    def delete_selected(self) -> bool:
        if self.is_read_only():
            return False
        self._remove_indices(self.selectedIndexes())
        return True

    def _remove_indices(self, indices: Iterable[QModelIndex]) -> None:
        # Need to be careful that the indexes are not invalidated by the removal of
        # previous rows, that's why we convert them to QPersistentModelIndex.
        persistent_indices = [QPersistentModelIndex(index) for index in indices]
        for index in persistent_indices:
            if index.isValid():
                self._model.removeRow(index.row(), index.parent())

    def copy_selection(self) -> None:
        selected_indexes = self.selectedIndexes()
        data = self._model.mimeData(selected_indexes)
        clipboard = QGuiApplication.clipboard()
        clipboard.setMimeData(data)

    def cut_selected(self) -> bool:
        if self.is_read_only():
            return False

        self.copy_selection()
        self._remove_indices(self.selectedIndexes())
        return True

    def paste(self) -> bool:
        if self.is_read_only():
            return False

        clipboard = QGuiApplication.clipboard()
        data = clipboard.mimeData()

        if self.selectedIndexes():
            first = QPersistentModelIndex(self.selectedIndexes()[0])
            parent = first.parent()
            row = first.row()
            if self._model.dropMimeData(data, Qt.DropAction.MoveAction, row, 0, parent):
                self._model.removeRow(first.row(), first.parent())
        else:
            parent = QModelIndex()
            row = self._model.rowCount()
            self._model.dropMimeData(data, Qt.DropAction.MoveAction, row, 0, parent)
        return True

    def show_context_menu(self, position):
        index = self.indexAt(position)
        if not index.isValid():
            return
        menu = QMenu(self)

        add_menu = QMenu()
        add_menu.setTitle("Insert above...")
        menu.addMenu(add_menu)

        create_variable_action = QAction("variable")
        add_menu.addAction(create_variable_action)
        new_variable = create_variable_declaration()
        create_variable_action.triggered.connect(
            functools.partial(self._model.insert_above, new_variable, index)
        )
        create_shot_action = QAction("Shot")
        add_menu.addAction(create_shot_action)
        new_shot = ExecuteShot()
        create_shot_action.triggered.connect(
            functools.partial(self._model.insert_above, new_shot, index)
        )
        create_linspace_action = QAction("Linspace loop")
        add_menu.addAction(create_linspace_action)
        new_linspace = create_linspace_loop()
        create_linspace_action.triggered.connect(
            functools.partial(self._model.insert_above, new_linspace, index)
        )
        create_arange_action = QAction("Arange loop")
        add_menu.addAction(create_arange_action)
        new_arange = create_arange_loop()
        create_arange_action.triggered.connect(
            functools.partial(self._model.insert_above, new_arange, index)
        )
        menu.exec(self.mapToGlobal(position))

    def copy_to_clipboard(self):
        steps = self.get_iteration()

        unstructured = serialization.converters["json"].unstructure(
            steps, StepsConfiguration
        )

        text = yaml.safe_dump(unstructured)
        clipboard = QGuiApplication.clipboard()
        clipboard.setText(text)

    def paste_from_clipboard(self) -> bool:
        """Attempts to paste the content of the clipboard.

        If the editor is in read-only mode, this method does nothing.
        """

        if self._model.is_read_only():
            return False

        clipboard = QGuiApplication.clipboard()
        text = clipboard.text()
        try:
            data = yaml.safe_load(text)
        except yaml.YAMLError as e:
            QtWidgets.QMessageBox.warning(
                self,
                "Invalid YAML content",
                f"Could not parse the clipboard content as YAML:\n {e}",
            )
            return False

        # TODO: raise recoverable exception if data is not valid
        steps = serialization.converters["json"].structure(data, StepsConfiguration)
        self.set_iteration(steps)

        return True



import attrs

from ._protocol import CondetrolExtensionProtocol
from ..device_configuration_editors._extension import CondetrolDeviceExtension
from ..timelanes_editor.extension import CondetrolLaneExtension


@attrs.define
class CondetrolExtension(CondetrolExtensionProtocol):
    lane_extension: CondetrolLaneExtension = attrs.field(factory=CondetrolLaneExtension)
    device_extension: CondetrolDeviceExtension = attrs.field(
        factory=CondetrolDeviceExtension
    )



from ._client import RPCClient
from ._configuration import (
    RPCConfiguration,
    InsecureRPCConfiguration,
)
from ._server import RPCServer, Server, RemoteError, RemoteCallError, InvalidProxyError
from .._proxy import Proxy


__all__ = [
    "Server",
    "Proxy",
    "RPCConfiguration",
    "InsecureRPCConfiguration",
    "RPCServer",
    "RPCClient",
    "RemoteError",
    "RemoteCallError",
    "InvalidProxyError",
]



from __future__ import annotations

from abc import abstractmethod
from datetime import datetime
from typing import Protocol, TYPE_CHECKING

from returns.result import Result

from .._path import PureSequencePath
from .._path_hierarchy import PathNotFoundError, PathIsRootError

if TYPE_CHECKING:
    from .._sequence_collection import PathIsSequenceError


class AsyncPathHierarchy(Protocol):
    @abstractmethod
    async def does_path_exists(self, path: PureSequencePath) -> bool:
        raise NotImplementedError

    @abstractmethod
    async def get_children(
        self, path: PureSequencePath
    ) -> Result[set[PureSequencePath], PathNotFoundError | PathIsSequenceError]:
        raise NotImplementedError

    @abstractmethod
    async def get_path_creation_date(
        self, path: PureSequencePath
    ) -> Result[datetime, PathNotFoundError | PathIsRootError]:
        raise NotImplementedError



from __future__ import annotations

import abc
import concurrent.futures
import logging
import threading
from collections.abc import Mapping
from contextlib import AbstractContextManager
from typing import Optional

import anyio
import anyio.from_thread

from caqtus.device import DeviceConfiguration, DeviceName
from caqtus.session import (
    ExperimentSessionMaker,
    PureSequencePath,
    ExperimentSession,
    State,
)
from caqtus.types.parameter import ParameterNamespace
from .._shot_compiler import create_shot_compiler
from .._shot_runner import create_shot_runner
from ..device_manager_extension import DeviceManagerExtensionProtocol
from ..sequence_runner import ShotRetryConfig
from ..sequence_runner.sequence_manager import run_sequence

logger = logging.getLogger(__name__)


class ExperimentManager(abc.ABC):
    @abc.abstractmethod
    def create_procedure(
        self, procedure_name: str, acquisition_timeout: Optional[float] = None
    ) -> Procedure:
        raise NotImplementedError

    @abc.abstractmethod
    def interrupt_running_procedure(self) -> bool:
        """Indicates to the active procedure that it must stop running sequences.

        Returns:
            True if there was an active procedure, and it signaled that it will stop
            running sequences.
            False if no procedure was active.
        """

        raise NotImplementedError


class Procedure(AbstractContextManager, abc.ABC):
    """Used to perform a procedure on the experiment.

    A procedure is anything more complex than a single sequence.
    It can be a sequence with some analysis performed afterward, a sequence that is run
    multiple times with different parameters, multiple sequences that must be run
    cohesively, etc...

    Procedures are created with :meth:`ExperimentManager.create_procedure`.

    The procedure must be active to start running sequences.
    A procedure is activated by using it as a context manager.
    No two procedures can be active at the same time.
    If a previous procedure is active, entering another procedure will block until the
    first procedure is exited.

    To run a sequence once a procedure is active, use :meth:`run_sequence`.

    Examples:

    .. code-block:: python

            experiment_manager: ExperimentManager = ...
            with experiment_manager.create_procedure("my procedure") as procedure:
                procedure.run_sequence(PureSequencePath("my sequence"))
                # do analysis, overwrite parameters, etc...
                procedure.run_sequence(PureSequencePath("another sequence"))
    """

    @abc.abstractmethod
    def is_active(self) -> bool:
        """Indicates if the procedure is currently active and can run sequences."""

        raise NotImplementedError

    @abc.abstractmethod
    def is_running_sequence(self) -> bool:
        """Indicates if the procedure is currently running a sequence."""

        raise NotImplementedError

    @abc.abstractmethod
    def exception(self) -> Optional[Exception]:
        """Retrieve the exception that occurred while running the last sequence.

        If a sequence is currently running, this method will block until the sequence
        is finished.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def start_sequence(
        self,
        sequence: PureSequencePath,
        global_parameters: Optional[ParameterNamespace] = None,
        device_configurations: Optional[
            Mapping[DeviceName, DeviceConfiguration]
        ] = None,
    ) -> None:
        """Start running the sequence on the setup.

        This method returns immediately, and the sequence is launched in a separate
        thread.

        Exceptions that occur while running the sequence are not raised by this method,
        but can be retrieved with the `exception` method.

        Args:
            sequence: the sequence to run.
            global_parameters: The parameters to set for this sequence.
            If nothing is passed, it will take the current global parameters from the
            session.
            device_configurations: the device configurations to use for running this
            sequence.
            If None, this will use the session default device configurations.
        Raises:
            ProcedureNotActiveError: if the procedure is not active.
            SequenceAlreadyRunningError: if a sequence is already running.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def interrupt_sequence(self) -> bool:
        """Interrupt the currently running sequence.

        This method only signals the sequence that it must interrupt as soon as
        possible, but it does not wait for the sequence to finish.
        To wait for the sequence to finish, use :meth:`wait_until_sequence_finished`
        after calling :meth:`interrupt_sequence`.

        Returns:
            True if a sequence was running and was interrupted.
            False if no sequence was running.
        """

        raise NotImplementedError

    def run_sequence(
        self,
        sequence: PureSequencePath,
        global_parameters: Optional[ParameterNamespace] = None,
        device_configurations: Optional[
            Mapping[DeviceName, DeviceConfiguration]
        ] = None,
    ) -> None:
        """Run a sequence on the setup.

        This method blocks until the sequence is finished.

        Arguments are the same as :meth:`start_sequence`.

        Raises:
            ProcedureNotActiveError: if the procedure is not active.
            SequenceAlreadyRunningError: if a sequence is already running.
            Exception: if an exception occurs while running the sequence.
        """

        self.start_sequence(sequence, global_parameters, device_configurations)
        if exception := self.exception():
            raise exception

    @abc.abstractmethod
    def sequences(self) -> list[PureSequencePath]:
        """Retrieve the list of sequences that were started by the procedure.

        Returns:
            A list of sequences that were started by the procedure since it was started,
            ordered by execution order.
            If the procedure is currently running a sequence, the sequence will be the
            last element of the list.
        """

        raise NotImplementedError


class LocalExperimentManager(ExperimentManager):
    """Implementation of :class:`ExperimentManager` that runs in the local process."""

    def __init__(
        self,
        session_maker: ExperimentSessionMaker,
        device_manager_extension: DeviceManagerExtensionProtocol,
        shot_retry_config: Optional[ShotRetryConfig] = None,
    ):
        self._procedure_running = threading.Lock()
        self._session_maker = session_maker
        self._shot_retry_config = shot_retry_config
        self._thread_pool = concurrent.futures.ThreadPoolExecutor(max_workers=1)
        self._active_procedure: Optional[BoundProcedure] = None
        self._device_manager_extension = device_manager_extension

        # We crash the sequences that might have been running previously.
        # This ensures that only one experiment manager is active at a time.
        # It also cleans up previous sequences that might still be running if the
        # previous experiment manager was not properly closed.
        self._crash_running_sequences()

    def _crash_running_sequences(self) -> None:
        with self._session_maker() as session:
            _crash_running_sequences(session)

    def __enter__(self):
        self._thread_pool.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        with self._procedure_running:
            return self._thread_pool.__exit__(exc_type, exc_value, traceback)

    def create_procedure(
        self, procedure_name: str, acquisition_timeout: Optional[float] = None
    ) -> BoundProcedure:
        return BoundProcedure(
            experiment_manager=self,
            name=procedure_name,
            session_maker=self._session_maker,
            lock=self._procedure_running,
            thread_pool=self._thread_pool,
            shot_retry_config=self._shot_retry_config,
            acquisition_timeout=acquisition_timeout,
            device_manager_extension=self._device_manager_extension,
        )

    def interrupt_running_procedure(self) -> bool:
        if self._active_procedure is None:
            return False
        return self._active_procedure.interrupt_sequence()


class BoundProcedure(Procedure):
    """Implementation of :class:`Procedure`.

    See :class:`Procedure` for documentation.

    This class is not meant to be instantiated directly, but is returned by
    :meth:`BoundExperimentManager.create_procedure`.
    """

    def __init__(
        self,
        experiment_manager: LocalExperimentManager,
        name: str,
        session_maker: ExperimentSessionMaker,
        lock: threading.Lock,
        thread_pool: concurrent.futures.ThreadPoolExecutor,
        shot_retry_config: ShotRetryConfig,
        device_manager_extension: DeviceManagerExtensionProtocol,
        acquisition_timeout: Optional[float] = None,
    ):
        self._parent = experiment_manager
        self._name = name
        self._session_maker = session_maker
        self._running = lock
        self._thread_pool = thread_pool
        self._sequence_future: Optional[concurrent.futures.Future] = None
        self._sequences: list[PureSequencePath] = []
        self._acquisition_timeout = acquisition_timeout if acquisition_timeout else -1
        self._shot_retry_config = shot_retry_config
        self._device_manager_extension = device_manager_extension
        self._cancel_scope: Optional[anyio.CancelScope] = None
        self._portal: Optional[anyio.from_thread.BlockingPortal] = None

    def __repr__(self):
        return f"<{self.__class__.__name__}('{self}') at {hex(id(self))}>"

    def __str__(self):
        return self._name

    def __enter__(self):
        if not self._running.acquire(timeout=self._acquisition_timeout):
            raise TimeoutError(f"Could not activate procedure <{self}>.")
        self._parent._active_procedure = self
        self._sequences.clear()
        return self

    def is_active(self) -> bool:
        return self._running.locked()

    def is_running_sequence(self) -> bool:
        return self._sequence_future is not None and not self._sequence_future.done()

    def sequences(self) -> list[PureSequencePath]:
        return self._sequences.copy()

    def exception(self) -> Optional[Exception]:
        if self._sequence_future is None:
            return None
        return self._sequence_future.exception()

    def start_sequence(
        self,
        sequence: PureSequencePath,
        global_parameters: Optional[ParameterNamespace] = None,
        device_configurations: Optional[
            Mapping[DeviceName, DeviceConfiguration]
        ] = None,
    ) -> None:
        if not self.is_active():
            exception = ProcedureNotActiveError("The procedure is not active.")
            exception.add_note(
                "It is only possible to run sequences inside active procedures."
            )
            exception.add_note(
                "Maybe you forgot to use the procedure inside a `with` statement?"
            )
            raise exception
        if self.is_running_sequence():
            raise SequenceAlreadyRunningError("A sequence is already running.")
        self._sequence_future = self._thread_pool.submit(
            self._run_sequence,
            sequence,
            global_parameters,
            device_configurations,
        )
        self._sequences.append(sequence)

    def interrupt_sequence(self) -> bool:
        if not self.is_running_sequence():
            return False
        assert self._cancel_scope is not None
        assert self._portal is not None
        self._portal.call(self._cancel_scope.cancel)
        return True

    def wait_until_sequence_finished(self):
        if self.is_running_sequence():
            self._sequence_future.result()

    def _run_sequence(
        self,
        sequence: PureSequencePath,
        global_parameters: Optional[ParameterNamespace] = None,
        device_configurations: Optional[
            Mapping[DeviceName, DeviceConfiguration]
        ] = None,
    ) -> None:
        async def run():
            with anyio.CancelScope() as self._cancel_scope:
                with anyio.from_thread.BlockingPortal() as self._portal:
                    await run_sequence(
                        sequence=sequence,
                        session_maker=self._session_maker,
                        shot_retry_config=self._shot_retry_config,
                        global_parameters=global_parameters,
                        device_configurations=device_configurations,
                        device_manager_extension=self._device_manager_extension,
                        shot_runner_factory=create_shot_runner,
                        shot_compiler_factory=create_shot_compiler,
                    )

        try:
            anyio.run(run, backend="trio")
        except Exception as e:
            logger.error(f"Error while running sequence {sequence}.", exc_info=e)
        self._cancel_scope = None

    def __exit__(self, exc_type, exc_value, traceback):
        error_occurred = exc_value is not None
        try:
            if error_occurred:
                self.interrupt_sequence()
            self.wait_until_sequence_finished()
        finally:
            self._parent._active_procedure = None
            self._running.release()


def _crash_running_sequences(session: ExperimentSession) -> None:
    for path in session.sequences.get_sequences_in_state(State.RUNNING):
        session.sequences.set_state(path, State.CRASHED)
    for path in session.sequences.get_sequences_in_state(State.PREPARING):
        session.sequences.set_state(path, State.CRASHED)


class SequenceAlreadyRunningError(RuntimeError):
    pass


class ProcedureNotActiveError(RuntimeError):
    pass



#!/usr/bin/python
from PySide6 import QtGui, QtCore, QtWidgets

from ..constants import ViewerEnum


class NodeScene(QtWidgets.QGraphicsScene):

    def __init__(self, parent=None):
        super(NodeScene, self).__init__(parent)
        self._grid_mode = ViewerEnum.GRID_DISPLAY_LINES.value
        self._grid_color = ViewerEnum.GRID_COLOR.value
        self._bg_color = ViewerEnum.BACKGROUND_COLOR.value
        self.setBackgroundBrush(QtGui.QColor(*self._bg_color))

    def __repr__(self):
        cls_name = str(self.__class__.__name__)
        return '<{}("{}") object at {}>'.format(cls_name, self.viewer(), hex(id(self)))

    # def _draw_text(self, painter, pen):
    #     font = QtGui.QFont()
    #     font.setPixelSize(48)
    #     painter.setFont(font)
    #     parent = self.viewer()
    #     pos = QtCore.QPoint(20, parent.height() - 20)
    #     painter.setPen(pen)
    #     painter.drawText(parent.mapToScene(pos), 'Not Editable')

    def _draw_grid(self, painter, rect, pen, grid_size):
        """
        draws the grid lines in the scene.

        Args:
            painter (QtGui.QPainter): painter object.
            rect (QtCore.QRectF): rect object.
            pen (QtGui.QPen): pen object.
            grid_size (int): grid size.
        """
        left = int(rect.left())
        right = int(rect.right())
        top = int(rect.top())
        bottom = int(rect.bottom())

        first_left = left - (left % grid_size)
        first_top = top - (top % grid_size)

        lines = []
        lines.extend(
            [
                QtCore.QLineF(x, top, x, bottom)
                for x in range(first_left, right, grid_size)
            ]
        )
        lines.extend(
            [
                QtCore.QLineF(left, y, right, y)
                for y in range(first_top, bottom, grid_size)
            ]
        )

        painter.setPen(pen)
        painter.drawLines(lines)

    def _draw_dots(self, painter, rect, pen, grid_size):
        """
        draws the grid dots in the scene.

        Args:
            painter (QtGui.QPainter): painter object.
            rect (QtCore.QRectF): rect object.
            pen (QtGui.QPen): pen object.
            grid_size (int): grid size.
        """
        zoom = self.viewer().get_zoom()
        if zoom < 0:
            grid_size = int(abs(zoom) / 0.3 + 1) * grid_size

        left = int(rect.left())
        right = int(rect.right())
        top = int(rect.top())
        bottom = int(rect.bottom())

        first_left = left - (left % grid_size)
        first_top = top - (top % grid_size)

        pen.setWidth(grid_size / 10)
        painter.setPen(pen)

        [
            painter.drawPoint(int(x), int(y))
            for x in range(first_left, right, grid_size)
            for y in range(first_top, bottom, grid_size)
        ]

    def drawBackground(self, painter, rect):
        super(NodeScene, self).drawBackground(painter, rect)

        painter.save()
        painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing, False)
        painter.setBrush(self.backgroundBrush())

        if self._grid_mode is ViewerEnum.GRID_DISPLAY_DOTS.value:
            pen = QtGui.QPen(QtGui.QColor(*self.grid_color), 0.65)
            self._draw_dots(painter, rect, pen, ViewerEnum.GRID_SIZE.value)

        elif self._grid_mode is ViewerEnum.GRID_DISPLAY_LINES.value:
            zoom = self.viewer().get_zoom()
            if zoom > -0.5:
                pen = QtGui.QPen(QtGui.QColor(*self.grid_color), 0.65)
                self._draw_grid(painter, rect, pen, ViewerEnum.GRID_SIZE.value)

            color = QtGui.QColor(*self._bg_color).darker(200)
            if zoom < -0.0:
                color = color.darker(100 - int(zoom * 110))
            pen = QtGui.QPen(color, 0.65)
            self._draw_grid(painter, rect, pen, ViewerEnum.GRID_SIZE.value * 8)

        painter.restore()

    def mousePressEvent(self, event):
        selected_nodes = self.viewer().selected_nodes()
        if self.viewer():
            self.viewer().sceneMousePressEvent(event)
        super(NodeScene, self).mousePressEvent(event)
        keep_selection = any(
            [
                event.button() == QtCore.Qt.MiddleButton,
                event.button() == QtCore.Qt.RightButton,
                event.modifiers() == QtCore.Qt.AltModifier,
            ]
        )
        if keep_selection:
            for node in selected_nodes:
                node.setSelected(True)

    def mouseMoveEvent(self, event):
        if self.viewer():
            self.viewer().sceneMouseMoveEvent(event)
        super(NodeScene, self).mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        if self.viewer():
            self.viewer().sceneMouseReleaseEvent(event)
        super(NodeScene, self).mouseReleaseEvent(event)

    def viewer(self):
        return self.views()[0] if self.views() else None

    @property
    def grid_mode(self):
        return self._grid_mode

    @grid_mode.setter
    def grid_mode(self, mode=None):
        if mode is None:
            mode = ViewerEnum.GRID_DISPLAY_LINES.value
        self._grid_mode = mode

    @property
    def grid_color(self):
        return self._grid_color

    @grid_color.setter
    def grid_color(self, color=(0, 0, 0)):
        self._grid_color = color

    @property
    def background_color(self):
        return self._bg_color

    @background_color.setter
    def background_color(self, color=(0, 0, 0)):
        self._bg_color = color
        self.setBackgroundBrush(QtGui.QColor(*self._bg_color))



import functools
from typing import TypeVar

from caqtus.types.iteration import (
    IterationConfiguration,
    StepsConfiguration,
)

from .sequence_iteration_editor import SequenceIterationEditor
from .steps_iteration_editor import StepsIterationEditor

T = TypeVar("T", bound=IterationConfiguration)


@functools.singledispatch
def create_default_editor(iteration: T) -> SequenceIterationEditor[T]:
    raise NotImplementedError


@create_default_editor.register
def _(iteration: StepsConfiguration) -> SequenceIterationEditor[StepsConfiguration]:
    return StepsIterationEditor(iteration)



"""Defines utility functions for assertions.

All functions defined in this module are only active when __debug__ is True, i.e. when
the program is run in debug mode.
When not in debug mode, the functions are no-ops.
"""

from collections.abc import Callable, MutableSequence
from typing import Concatenate, TypeVar, ParamSpec

T = TypeVar("T")
P = ParamSpec("P")
S = TypeVar("S", bound=MutableSequence)


def assert_length_changed(
    difference: int,
) -> Callable[[Callable[Concatenate[S, P], T]], Callable[Concatenate[S, P], T]]:
    """Decorator for functions that change the length of a sequence.

    The first argument of the function must be a mutable sequence.
    If the length of the sequence does not change by the expected amount, an
    AssertionError is raised.

    Args:
        difference: The expected difference in length before and after the function is
        called.
    """

    def decorator(
        fun: Callable[Concatenate[S, P], T]
    ) -> Callable[Concatenate[S, P], T]:
        if __debug__:

            def wrapper(self, *args, **kwargs):
                length_before = len(self)
                result = fun(self, *args, **kwargs)
                length_after = len(self)
                assert length_after == length_before + difference
                return result

            return wrapper
        else:
            return fun

    return decorator



from ._delegate import TimeLaneDelegate
from .digital_lane_model import DigitalTimeLaneModel
from .analog_lane_model import AnalogTimeLaneModel
from .model import TimeLanesModel, TimeLaneModel
from .time_lanes_editor import TimeLanesEditor

__all__ = [
    "TimeLanesEditor",
    "TimeLanesModel",
    "DigitalTimeLaneModel",
    "TimeLaneModel",
    "TimeLaneDelegate",
    "AnalogTimeLaneModel",
]



from datetime import datetime
from datetime import timezone
from typing import TYPE_CHECKING

import sqlalchemy.orm
from attr import frozen
from returns.result import Success, Failure, Result
from sqlalchemy import select
from sqlalchemy.orm import Session

from ._path_table import SQLSequencePath
from .._path import PureSequencePath
from .._path_hierarchy import (
    PathNotFoundError,
    PathIsRootError,
    PathHierarchy,
)
from .._return_or_raise import unwrap, is_success
from .._sequence_collection import PathIsSequenceError

if TYPE_CHECKING:
    from ._experiment_session import SQLExperimentSession


@frozen
class SQLPathHierarchy(PathHierarchy):
    parent_session: "SQLExperimentSession"

    def does_path_exists(self, path: PureSequencePath) -> bool:
        return _does_path_exists(self._get_sql_session(), path)

    def create_path(
        self, path: PureSequencePath
    ) -> Result[list[PureSequencePath], PathIsSequenceError]:
        paths_to_create: list[PureSequencePath] = []
        current = path
        sequence_collection = self.parent_session.sequences
        while parent := current.parent:
            match sequence_collection.is_sequence(current):
                case Success(True):
                    return Failure(
                        PathIsSequenceError(
                            f"Cannot create path {path} because {current} is already a"
                            " sequence"
                        )
                    )
                case Success(False):
                    pass
                case Failure(PathNotFoundError()):
                    paths_to_create.append(current)
            current = parent

        session = self._get_sql_session()
        created_paths = []
        for path_to_create in reversed(paths_to_create):
            parent = (
                unwrap(self._query_path_model(path_to_create.parent))
                if not path_to_create.parent.is_root()
                else None
            )
            new_path = SQLSequencePath(
                path=str(path_to_create),
                parent=parent,
                creation_date=datetime.now(tz=timezone.utc).replace(tzinfo=None),
            )
            session.add(new_path)
            created_paths.append(path_to_create)
        return Success(created_paths)

    def get_children(
        self, path: PureSequencePath
    ) -> Result[set[PureSequencePath], PathNotFoundError | PathIsSequenceError]:
        return _get_children(self._get_sql_session(), path)

    def delete_path(self, path: PureSequencePath, delete_sequences: bool = False):
        session = self._get_sql_session()

        if not delete_sequences:
            sequence_collection = self.parent_session.sequences
            if contained := sequence_collection.get_contained_sequences(path):
                raise PathIsSequenceError(
                    f"Cannot delete a path that contains sequences: {contained}"
                )

        session.delete(unwrap(self._query_path_model(path)))
        session.flush()

    def get_all_paths(self) -> set[PureSequencePath]:
        query = select(SQLSequencePath)
        result = self._get_sql_session().execute(query)
        return {PureSequencePath(path.path) for path in result.scalars()}

    def update_creation_date(self, path: PureSequencePath, date: datetime) -> None:
        if path.is_root():
            raise PathIsRootError(path)

        sql_path = unwrap(self._query_path_model(path))
        sql_path.creation_date = date

    def _query_path_model(
        self, path: PureSequencePath
    ) -> Result[SQLSequencePath, PathNotFoundError]:
        return _query_path_model(self._get_sql_session(), path)

    def _get_sql_session(self) -> sqlalchemy.orm.Session:
        # noinspection PyProtectedMember
        return self.parent_session._get_sql_session()

    def get_path_creation_date(
        self, path: PureSequencePath
    ) -> Result[datetime, PathNotFoundError]:
        return _get_path_creation_date(self._get_sql_session(), path)


def _does_path_exists(session: Session, path: PureSequencePath) -> bool:
    if path.is_root():
        return True
    result = _query_path_model(session, path)
    return is_success(result)


def _get_children(
    session: Session, path: PureSequencePath
) -> Result[set[PureSequencePath], PathNotFoundError | PathIsSequenceError]:
    if path.is_root():
        query_children = select(SQLSequencePath).where(
            SQLSequencePath.parent_id.is_(None)
        )
        children = session.scalars(query_children)
    else:
        query_result = _query_path_model(session, path)
        match query_result:
            case Success(path_sql):
                if path_sql.sequence:
                    return Failure(
                        PathIsSequenceError(
                            f"Cannot check children of a sequence: {path}"
                        )
                    )
                else:
                    children = path_sql.children
            case Failure() as failure:
                return failure
    # noinspection PyUnboundLocalVariable
    return Success(set(PureSequencePath(str(child.path)) for child in children))


def _get_path_creation_date(
    session: Session, path: PureSequencePath
) -> Result[datetime, PathNotFoundError]:
    if path.is_root():
        return Failure(PathIsRootError(path))
    return _query_path_model(session, path).map(
        lambda x: x.creation_date.replace(tzinfo=timezone.utc)
    )


def _query_path_model(
    session: Session, path: PureSequencePath
) -> Result[SQLSequencePath, PathNotFoundError]:
    stmt = select(SQLSequencePath).where(SQLSequencePath.path == str(path))
    result = session.execute(stmt)
    if found := result.scalar():
        return Success(found)
    else:
        return Failure(PathNotFoundError(f'Path "{path}" does not exists'))



from collections.abc import Iterable
from itertools import islice
from typing import TypeVar

T = TypeVar("T")


def batched(iterable: Iterable[T], n: int) -> Iterable[tuple[T, ...]]:
    """Batch data into tuples of length n. The last batch may be shorter."""
    # batched('ABCDEFG', 3) --> ABC DEF G

    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := tuple(islice(it, n)):
        yield batch



#!/usr/bin/python
# -*- coding: utf-8 -*-
__version__ = "0.6.37"
__status__ = "Work in Progress"
__license__ = "MIT"

__author__ = "Johnny Chan"

__module_name__ = "NodeGraphQt"
__url__ = "https://github.com/jchanvfx/NodeGraphQt"



from __future__ import annotations

import datetime
from typing import Optional

import sqlalchemy
from sqlalchemy import ForeignKey, DateTime, UniqueConstraint, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from ._path_table import SQLSequencePath
from ._shot_tables import SQLShot
from ._table_base import Base
from .._state import State


class SQLSequence(Base):
    """This table stores the sequences that are created by the user.

    Attributes:
        id_: The unique identifier of the sequence.
        path_id: The identifier of the path that the sequence is associated with.
        path: A reference to the path that the sequence is associated with.
        state: The state of the sequence.
        exception_traceback: The traceback of the exception that occurred while running
            the sequence.

            This is None if the sequence is not in the CRASHED state.

            It can be None even if the sequence is in the CRASHED state if the traceback
            was not captured.

        parameters: A reference to the parameters of the sequence.
        iteration: A reference to the iteration configuration of the sequence.
        time_lanes: A reference to the time lanes of the sequence.
        device_configurations: A reference to the device configurations of the sequence.

        start_time: The time at which the sequence started execution.

            Stored as a timezone naive datetime, with the assumption that the timezone
            is UTC.

        stop_time: The time at which the sequence stopped execution.

            Stored as a timezone naive datetime, with the assumption that the timezone
            is UTC.

        shots: A reference to the shots that are part of the sequence.
        expected_number_of_shots: The number of shots that are expected to be executed
            in total for this sequence, inferred from the iteration configuration.

            Can be None if this value is not known.
    """

    __tablename__ = "sequences"

    id_: Mapped[int] = mapped_column(name="id", primary_key=True)
    path_id: Mapped[str] = mapped_column(
        ForeignKey(SQLSequencePath.id_, ondelete="CASCADE"), unique=True, index=True
    )
    path: Mapped[SQLSequencePath] = relationship(back_populates="sequence")
    state: Mapped[State]
    exception_traceback: Mapped[Optional[SQLExceptionTraceback]] = relationship(
        cascade="all, delete", passive_deletes=True, back_populates="sequence"
    )

    parameters: Mapped[SQLSequenceParameters] = relationship(
        cascade="all, delete",
        passive_deletes=True,
        back_populates="sequence",
    )
    iteration: Mapped[SQLIterationConfiguration] = relationship(
        cascade="all",
        passive_deletes=True,
        back_populates="sequence",
    )
    time_lanes: Mapped[SQLTimelanes] = relationship(
        cascade="all, delete",
        passive_deletes=True,
        back_populates="sequence",
    )

    device_configurations: Mapped[list[SQLDeviceConfiguration]] = relationship(
        cascade="all, delete", passive_deletes=True, back_populates="sequence"
    )

    # Stored as timezone naive datetimes, with the assumption that the timezone is UTC.
    start_time: Mapped[Optional[datetime.datetime]] = mapped_column(
        DateTime(timezone=False)
    )
    stop_time: Mapped[Optional[datetime.datetime]] = mapped_column(
        DateTime(timezone=False)
    )

    shots: Mapped[list["SQLShot"]] = relationship(
        back_populates="sequence",
        cascade="all, delete",
        passive_deletes=True,
    )
    # expected_number_of_shots indicates how many shots are expected to be executed in
    # total for this sequence.
    # It is written when the sequence iteration is set.
    # None indicates that this value is not known.
    expected_number_of_shots: Mapped[Optional[int]] = mapped_column()


class SQLSequenceParameters(Base):
    __tablename__ = "sequence.parameters"

    id_: Mapped[int] = mapped_column(name="id", primary_key=True)
    sequence_id: Mapped[int] = mapped_column(
        ForeignKey(SQLSequence.id_, ondelete="CASCADE"), index=True
    )
    sequence: Mapped[SQLSequence] = relationship(
        back_populates="parameters", single_parent=True
    )
    content = mapped_column(sqlalchemy.types.JSON)

    __table_args__ = (UniqueConstraint(sequence_id),)


class SQLIterationConfiguration(Base):
    __tablename__ = "sequence.iteration"

    id_: Mapped[int] = mapped_column(name="id", primary_key=True)
    sequence_id: Mapped[int] = mapped_column(
        ForeignKey(SQLSequence.id_, ondelete="CASCADE"), index=True
    )
    sequence: Mapped[SQLSequence] = relationship(
        back_populates="iteration", single_parent=True
    )
    content = mapped_column(sqlalchemy.types.JSON)

    __table_args__ = (UniqueConstraint(sequence_id),)


class SQLTimelanes(Base):
    __tablename__ = "sequence.time_lanes"

    id_: Mapped[int] = mapped_column(name="id", primary_key=True)
    sequence_id: Mapped[int] = mapped_column(
        ForeignKey(SQLSequence.id_, ondelete="CASCADE"), index=True
    )
    sequence: Mapped[SQLSequence] = relationship(
        back_populates="time_lanes", single_parent=True
    )
    content = mapped_column(sqlalchemy.types.JSON)

    __table_args__ = (UniqueConstraint(sequence_id),)


class SQLDeviceConfiguration(Base):
    __tablename__ = "sequence.device_configurations"

    id_: Mapped[int] = mapped_column(primary_key=True)
    sequence_id: Mapped[int] = mapped_column(
        ForeignKey(SQLSequence.id_, ondelete="CASCADE")
    )
    sequence: Mapped[SQLSequence] = relationship(back_populates="device_configurations")
    name: Mapped[str] = mapped_column(String(255))
    device_type: Mapped[str] = mapped_column(String(255))
    content = mapped_column(sqlalchemy.types.JSON)

    __table_args__ = (
        sqlalchemy.UniqueConstraint(sequence_id, name, name="device_configuration"),
    )


class SQLExceptionTraceback(Base):
    __tablename__ = "sequence.exception_tracebacks"

    id_: Mapped[int] = mapped_column(name="id", primary_key=True)
    sequence_id: Mapped[int] = mapped_column(
        ForeignKey(SQLSequence.id_, ondelete="CASCADE"), index=True
    )
    sequence: Mapped[SQLSequence] = relationship(
        back_populates="exception_traceback", single_parent=True
    )
    content = mapped_column(sqlalchemy.types.JSON, nullable=False)

    __table_args__ = (UniqueConstraint(sequence_id),)



"""Defines region of interest to select a subset of an image."""

from caqtus.utils import serialization
from ._arbitrary_roi import ArbitraryROI
from ._converter import converter
from ._rectangular_roi import RectangularROI
from ._roi import ROI, Width, Height
from ._rotated_rectangular_roi import RotatedRectangularROI

# TODO: Remove this once nothing relies on the serialization module
serialization.include_subclasses(ROI)


__all__ = [
    "ROI",
    "RectangularROI",
    "RotatedRectangularROI",
    "ArbitraryROI",
    "Width",
    "Height",
    "converter",
]



from __future__ import annotations

from collections.abc import Mapping, Iterable, Sequence
from typing import Any, Optional

import attrs
import cattrs
import numpy as np

import caqtus.formatter as fmt
from caqtus.types.parameter import (
    magnitude_in_unit,
    add_unit,
    AnalogValue,
)
from caqtus.types.recoverable_exceptions import NotDefinedUnitError
from caqtus.types.units import (
    Unit,
    Quantity,
    UnitLike,
    DimensionalityError,
    InvalidDimensionalityError,
)
from caqtus.types.variable_name import DottedVariableName
from ._converter import converter, structure_evaluable_output
from .transformation import (
    Transformation,
    EvaluableOutput,
    evaluable_output_validator,
    OutputValue,
    evaluate,
)


def _data_points_converter(data_points: Iterable[tuple[float, float]]):
    point_to_tuple = [(x, y) for x, y in data_points]
    return tuple(sorted(point_to_tuple))


@attrs.define
class LinearInterpolation(Transformation):
    """Transforms an input value by applying a piecewise linear interpolation.

    This transformation stores a set of measured data points and interpolates the
    output value based on the input value.

    Attributes:
        input_: An operation that can be evaluated to an output value.
            The transformation is applied to this value.
        measured_data_points: A list of measured data points as tuples of input and
            output values.
        input_points_unit: The unit of the input points.
            The result of the input evaluation will be converted to this unit.
        output_points_unit: The unit of the output points.
            The result of the transformation will be converted to this unit.
    """

    input_: EvaluableOutput = attrs.field(
        validator=evaluable_output_validator,
        on_setattr=attrs.setters.validate,
    )
    measured_data_points: tuple[tuple[float, float], ...] = attrs.field(
        converter=_data_points_converter, on_setattr=attrs.setters.convert
    )
    input_points_unit: Optional[str] = attrs.field(
        converter=attrs.converters.optional(str),
        on_setattr=attrs.setters.convert,
    )
    output_points_unit: Optional[str] = attrs.field(
        converter=attrs.converters.optional(str),
        on_setattr=attrs.setters.convert,
    )

    def evaluate(self, variables: Mapping[DottedVariableName, Any]) -> OutputValue:
        input_value = evaluate(self.input_, variables)
        interpolator = Interpolator(
            self.measured_data_points, self.input_points_unit, self.output_points_unit
        )
        try:
            return interpolator(input_value)
        except DimensionalityError as e:
            raise InvalidDimensionalityError(f"Invalid dimensionality") from e


# Workaround for https://github.com/python-attrs/cattrs/issues/430
structure_hook = cattrs.gen.make_dict_structure_fn(
    LinearInterpolation,
    converter,
    input_=cattrs.override(struct_hook=structure_evaluable_output),
)

converter.register_structure_hook(LinearInterpolation, structure_hook)


def to_base_units(
    values: Sequence[float], required_unit: Optional[UnitLike]
) -> tuple[Sequence[float], Optional[Unit]]:

    if required_unit is None:
        return values, None
    try:
        unit = Unit(required_unit)
    except NotDefinedUnitError:
        raise NotDefinedUnitError(f"Undefined {fmt.unit(required_unit)}.")
    base_unit = Quantity(1, unit).to_base_units().units
    return [
        Quantity(value, unit).to(base_unit).magnitude for value in values
    ], base_unit


class Interpolator:
    def __init__(
        self,
        measured_data_points: Iterable[tuple[float, float]],
        input_units: Optional[UnitLike],
        output_units: Optional[UnitLike],
    ):
        measured_data_points = sorted(measured_data_points, key=lambda x: x[0])
        self.input_points, self.input_unit = to_base_units(
            [point[0] for point in measured_data_points], input_units
        )
        self.output_points, self.output_unit = to_base_units(
            [point[1] for point in measured_data_points], output_units
        )

    def __call__(self, input_value: AnalogValue) -> AnalogValue:

        input_value = magnitude_in_unit(input_value, self.input_unit)

        output_magnitude = np.interp(
            x=input_value,
            xp=self.input_points,
            fp=self.output_points,
            left=self.output_points[0],
            right=self.output_points[-1],
        )
        return add_unit(output_magnitude, self.output_unit)



import importlib.resources

import pint._typing

from caqtus.types.recoverable_exceptions import InvalidValueError

units_definition_file = importlib.resources.files("caqtus.types.units").joinpath(
    "units_definition.txt"
)

ureg = pint.UnitRegistry(
    units_definition_file,
    autoconvert_offset_to_baseunit=True,
    cache_folder=":auto:",
)
unit_registry = ureg
pint.set_application_registry(unit_registry)
Quantity = pint.Quantity
Unit = pint.Unit
UnitLike = pint._typing.UnitLike
UndefinedUnitError = pint.UndefinedUnitError

DimensionalityError = pint.DimensionalityError
dimensionless = Quantity(1, "").units

TIME_UNITS = {"s", "ms", "µs", "us", "ns"}

FREQUENCY_UNITS = {
    "Hz",
    "kHz",
    "MHz",
    "GHz",
    "THz",
}

POWER_UNITS = {
    "W",
    "mW",
    "dBm",
}

DIMENSIONLESS_UNITS = {"dB"}

CURRENT_UNITS = {"A", "mA"}

VOLTAGE_UNITS = {"V", "mV"}

DISTANCE_UNITS = {"m", "mm", "µm", "um", "nm"}

DEGREE_UNITS = {"deg", "rad"}

UNITS = (
    TIME_UNITS
    | FREQUENCY_UNITS
    | POWER_UNITS
    | DIMENSIONLESS_UNITS
    | CURRENT_UNITS
    | VOLTAGE_UNITS
    | DISTANCE_UNITS
    | DEGREE_UNITS
)


class InvalidDimensionalityError(InvalidValueError):
    """Raised when a value has an invalid dimensionality.

    This error is raised when a value has an invalid dimensionality and the user
    should fix it.
    """

    pass



import contextlib
from typing import TypeVar

import eliot

from caqtus.device.remote import DeviceProxy
from ._runtime import Camera

CameraType = TypeVar("CameraType", bound=Camera)


class CameraProxy(DeviceProxy[CameraType]):
    async def update_parameters(self, timeout: float, *args, **kwargs) -> None:
        return await self.call_method("update_parameters", timeout, *args, **kwargs)

    @contextlib.asynccontextmanager
    async def acquire(self, exposures: list[float]):
        async with (
            self.call_method_proxy_result("acquire", exposures) as cm_proxy,
            self.async_context_manager(cm_proxy) as iterator_proxy,
        ):
            yield self.async_iterator(iterator_proxy)
            return



"""This module defines recoverable exceptions.

These are exceptions that can occur during the normal execution of a program, for
example when the user enters an invalid value, or when a connection to an external
resource fails.

These errors will not crash caqtus applications, but instead will be caught and
displayed to the user, so that they can fix the error and retry the operation.

Exceptions that are not recoverable will not be caught, and are allowed to crash
the applications.

Only make exceptions recoverable if you expect them to happen in normal operation.
"""

from __future__ import annotations

import tblib.pickling_support


def is_recoverable(error: BaseException) -> bool:
    """Check if an error is recoverable.

    An error is recoverable if any of the following conditions are met:

    * The error is an instance of :class:`RecoverableException`.
    * The error's cause is a recoverable error.
    * The error is an instance of :class:`BaseExceptionGroup` and all its
      sub-exceptions are recoverable.

    Note that an error can be recoverable even if its cause is not recoverable, if
    the error itself is recoverable.
    """

    if isinstance(error, RecoverableException):
        return True

    if error.__cause__ is not None:
        return is_recoverable(error.__cause__)

    if isinstance(error, BaseExceptionGroup):
        return all(is_recoverable(e) for e in error.exceptions)

    return False


def split_recoverable(
    exception: BaseException,
) -> (
    tuple[BaseException, BaseException]
    | tuple[None, BaseException]
    | tuple[BaseException, None]
):
    """Split an exception into recoverable and non-recoverable parts.

    This function is mainly meant to split exception groups.

    Returns:
        A tuple of two elements:

        * The recoverable part of the exception, or None if there is no recoverable
          part.
        * The non-recoverable part of the exception, or None if there is no
          non-recoverable part.
    """

    if isinstance(exception, BaseExceptionGroup):
        return exception.split(is_recoverable)
    else:
        if is_recoverable(exception):
            return exception, None
        else:
            return None, exception


class RecoverableException(Exception):
    """An error that can be recovered from.

    This is an error that happen when the user does something wrong, and it is possible
    to retry the operation after they fix the error.

    This is a base class for all recoverable errors.
    It should not be raised directly, instead raise a subclass.
    """

    pass


class InvalidTypeError(TypeError, RecoverableException):
    """Raised when a value is not of the expected type.

    This error is raised when a value is not of the expected type, but it is possible
    to recover from the error by changing the value to the correct type.
    """

    pass


class InvalidValueError(ValueError, RecoverableException):
    """Raised when a value is invalid.

    This error is raised when a value is invalid, but it is possible to recover from the
    error by changing the value to a valid one.
    """

    pass


class ConnectionFailedError(ConnectionError, RecoverableException):
    """Raised when a connection to an external resource fails.

    This error is raised when a connection to an external resource fails, but it is
    possible to recover from the error by retrying the connection or fixing the
    connection settings.
    """

    pass


class ShotAttemptsExceededError(RecoverableException, ExceptionGroup):
    """Raised when the number of shot attempts exceeds the maximum allowed."""

    def derive(self, excs):
        return ShotAttemptsExceededError(self.message, excs)


class EvaluationError(RecoverableException):
    """Raised when an error occurs during the evaluation of an expression.

    This error is raised when an error occurs during the evaluation of an expression
    entered by the user.
    The user should fix the expression and retry the operation.
    """

    pass


class NotDefinedUnitError(RecoverableException):
    """Raised when the user tries to use a unit that is not defined."""


tblib.pickling_support.install(
    RecoverableException,
    InvalidTypeError,
    InvalidValueError,
    ConnectionFailedError,
    ShotAttemptsExceededError,
    EvaluationError,
    NotDefinedUnitError,
)



import abc
import copy
from typing import Optional

from PySide6.QtWidgets import QWidget, QFormLayout, QLineEdit

import caqtus.gui.qtutil.qabc as qabc
from caqtus.device import DeviceConfiguration
from caqtus.device.configuration import DeviceServerName


class DeviceConfigurationEditor[T: DeviceConfiguration](
    QWidget, metaclass=qabc.QABCMeta
):
    """A widget that allows to edit the configuration of a device.

    This class is generic in the type of the device configuration it allows to edit.
    """

    @abc.abstractmethod
    def get_configuration(self) -> T:
        """Return a new configuration that represents what is currently displayed."""

        raise NotImplementedError


class FormDeviceConfigurationEditor[T: DeviceConfiguration](
    DeviceConfigurationEditor[T]
):
    """Displays a list of fields to edit the configuration of a device.

    Widgets of this class initially only present a single field to edit the remote
    server name.

    Other device specific fields can be added by calling the :meth:`insert_row` method.
    """

    def __init__(self, device_configuration: T, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self.form = QFormLayout()
        self.device_configuration = copy.deepcopy(device_configuration)
        self.remote_server_line_edit = QLineEdit(self)
        self.remote_server_line_edit.setPlaceholderText("None")
        self.set_remote_server(self.device_configuration.remote_server)
        self.form.addRow("Remote server", self.remote_server_line_edit)
        self.setLayout(self.form)

    def append_row(self, label: str, widget: QWidget):
        """Append a widget field at the end of the form."""

        self.form.addRow(label, widget)

    def insert_row(self, label: str, widget: QWidget, row: int):
        """Insert a widget field at the specified row."""

        self.form.insertRow(row, label, widget)

    def get_configuration(self) -> T:
        """Return a new configuration with fields updated from the UI.

        Returns:
            A copy of the configuration that was passed to the constructor with the
            remote server field updated to the value set in the UI.

            Subclasses should override this method to update other fields as well.
        """

        configuration = copy.deepcopy(self.device_configuration)
        configuration.remote_server = self.read_remote_server()
        return configuration

    def set_remote_server(self, remote_server: Optional[DeviceServerName]) -> None:
        """Set the remote server name in the editor."""

        self.remote_server_line_edit.setText(remote_server or "")

    def read_remote_server(self) -> Optional[DeviceServerName]:
        """Read the remote server name from the editor."""

        text = self.remote_server_line_edit.text()
        if text == "":
            return None
        return DeviceServerName(text)



from typing import Optional, Any, assert_never

from PySide6.QtCore import QObject, QModelIndex, Qt
from PySide6.QtGui import QAction, QBrush, QPalette
from PySide6.QtWidgets import QMenu

from caqtus.types.expression import Expression
from caqtus.types.timelane import DigitalTimeLane
from .model import ColoredTimeLaneModel


class DigitalTimeLaneModel(ColoredTimeLaneModel[DigitalTimeLane, None]):
    def __init__(self, name: str, parent: Optional[QObject] = None):
        lane = DigitalTimeLane([False])
        super().__init__(name, lane, parent)
        if self._brush is None:
            # If no brush is set the button will be invisible, so we pick the
            # base color from the palette as default.
            color = QPalette().text().color()
            self._brush = QBrush(color)

    def data(
        self, index: QModelIndex, role: Qt.ItemDataRole = Qt.ItemDataRole.DisplayRole
    ):
        if not index.isValid():
            return None
        value = self._lane[index.row()]
        if role == Qt.ItemDataRole.DisplayRole:
            if isinstance(value, bool):
                return
            elif isinstance(value, Expression):
                return str(value)
            else:
                assert_never(value)
        elif role == Qt.ItemDataRole.EditRole:
            return value
        elif role == Qt.ItemDataRole.BackgroundRole:
            if isinstance(value, bool):
                if value:
                    return self._brush
        else:
            return super().data(index, role)

    def setData(
        self, index: QModelIndex, value: Any, role: int = Qt.ItemDataRole.EditRole
    ):
        if not index.isValid():
            return False
        if role == Qt.ItemDataRole.EditRole:
            start, stop = self._lane.get_bounds(index.row())
            if isinstance(value, bool):
                self._lane[start:stop] = value
                self.dataChanged.emit(index, index)
                return True
            elif isinstance(value, Expression):
                self._lane[start:stop] = value
                self.dataChanged.emit(index, index)
                return True
            else:
                raise TypeError(f"Invalid type for value: {type(value)}")
        return False

    def insertRow(self, row, parent: QModelIndex = QModelIndex()) -> bool:
        return self.insert_value(row, False)

    def get_cell_context_actions(self, index: QModelIndex) -> list[QAction | QMenu]:
        if not index.isValid():
            return []
        cell_type_menu = QMenu("Cell type")
        value = self._lane[index.row()]
        bool_action = cell_type_menu.addAction("on/off")
        if isinstance(value, bool):
            bool_action.setCheckable(True)
            bool_action.setChecked(True)
        else:
            bool_action.triggered.connect(
                lambda: self.setData(index, False, Qt.ItemDataRole.EditRole)
            )
        expr_action = cell_type_menu.addAction("expression")
        if isinstance(value, Expression):
            expr_action.setCheckable(True)
            expr_action.setChecked(True)
        else:
            expr_action.triggered.connect(
                lambda: self.setData(index, Expression("..."), Qt.ItemDataRole.EditRole)
            )

        return super().get_cell_context_actions(index) + [cell_type_menu]



from __future__ import annotations

import abc
from collections.abc import Set
from typing import Optional, assert_never, Literal

import anyio
import attrs
from PySide6.QtCore import Signal, Qt
from PySide6.QtGui import QIcon, QColor, QPalette, QKeySequence
from PySide6.QtWidgets import (
    QWidget,
    QToolBar,
    QStackedWidget,
    QLabel,
    QHBoxLayout,
    QApplication,
)

from caqtus.session import (
    ExperimentSessionMaker,
    PureSequencePath,
    ExperimentSession,
    AsyncExperimentSession,
    TracebackSummary,
)
from caqtus.session import (
    PathNotFoundError,
    SequenceNotEditableError,
    PathIsNotSequenceError,
    State,
)
from caqtus.session._return_or_raise import unwrap
from caqtus.types.iteration import (
    IterationConfiguration,
    StepsConfiguration,
)
from caqtus.types.parameter import ParameterNamespace
from caqtus.types.timelane import TimeLanes
from caqtus.types.variable_name import DottedVariableName
from .sequence_widget_ui import Ui_SequenceWidget
from .._icons import get_icon
from .._parameter_tables_editor import ParameterNamespaceEditor
from .._sequence_iteration_editors import StepsIterationEditor
from ..timelanes_editor import TimeLanesEditor
from ..timelanes_editor.extension import CondetrolLaneExtensionProtocol
from ..._common.exception_tree import ExceptionDialog


class SequenceWidget(QWidget, Ui_SequenceWidget):
    """Widget for editing sequence parameters, iterations and time lanes.

    This widget is a tab widget with three tabs: one for defining initial parameters, one for editing how the
    parameters should be iterated over for the sequence, and one for editing the time lanes that specify how
    a given shot should be executed.

    This widget is (optionally) associated with a sequence and displays the iteration
    configuration and time lanes for that sequence.
    If the widget is not associated with a sequence, it will hide itself.

    When associated with a sequence, the widget is constantly watching the state of the
    sequence.
    If the sequence is not in the draft state, the iteration editor and time lanes editor
    will become read-only.
    If the sequence is in the draft state, the iteration editor and time lanes editor
    will become editable and any change will be saved.
    """

    sequence_start_requested = Signal(PureSequencePath)
    """A signal emitted when the start sequence button is clicked."""

    def __init__(
        self,
        session_maker: ExperimentSessionMaker,
        extension: CondetrolLaneExtensionProtocol,
        parent: Optional[QWidget] = None,
    ):
        """Initializes the sequence widget.

        The sequence widget will initially be associated with no sequence.

        Args:
            session_maker: It is used to connect to the storage system in which to look
            for the sequence.
            extension: The extension that provides time lanes customization.
            parent: The parent widget.
        """

        super().__init__(parent)
        self.setupUi(self)
        self.session_maker = session_maker

        with self.session_maker() as session:
            device_configurations = dict(session.default_device_configurations)

        self._state: _State = _SequenceNotSetState()
        self._state_lock = anyio.Lock()

        self.parameters_editor = ParameterNamespaceEditor(self)
        self.parameters_editor.set_read_only(True)
        self.time_lanes_editor = TimeLanesEditor(
            extension,
            device_configurations,
            self,
        )
        self.iteration_editor = StepsIterationEditor(self)

        self.tabWidget.clear()
        self.tabWidget.addTab(self.parameters_editor, "&Globals")
        self.tabWidget.addTab(self.iteration_editor, "&Parameters")
        self.tabWidget.addTab(self.time_lanes_editor, "Time &lanes")

        self.setup_connections()

        self.tool_bar = QToolBar(self)
        self.status_widget = IconLabel(icon_position="left")

        self.warning_action = self.tool_bar.addAction(
            get_icon("mdi6.alert", color=QColor(205, 22, 17)), "warning"
        )
        self.warning_action.triggered.connect(self._on_warning_action_triggered)
        self.warning_action.setToolTip("Error")

        self.tool_bar.addWidget(self.status_widget)
        self.start_sequence_action = self.tool_bar.addAction(
            get_icon("start", color=Qt.GlobalColor.darkGreen), "start"
        )
        self.start_sequence_action.triggered.connect(self._on_start_sequence_requested)
        self.start_sequence_action.setShortcut(QKeySequence("F5"))
        self.interrupt_sequence_action = self.tool_bar.addAction(
            get_icon("stop"), "interrupt"
        )
        self.interrupt_sequence_action.setVisible(False)
        self.verticalLayout.insertWidget(0, self.tool_bar)
        self.stacked = QStackedWidget(self)
        self.stacked.addWidget(QWidget(self))
        self.stacked.addWidget(self.iteration_editor.toolbar)
        self.stacked.addWidget(self.time_lanes_editor.toolbar)
        self.stacked.setCurrentIndex(0)
        self.tool_bar.addSeparator()
        self.tool_bar.addWidget(self.stacked)

        self.tabWidget.currentChanged.connect(self.stacked.setCurrentIndex)

        self._transition(_SequenceNotSetState())

        self._exception_dialog = ExceptionDialog(self)

    def set_available_parameter_names(
        self, parameter_names: Set[DottedVariableName]
    ) -> None:
        """Set the names of the parameters that are defined externally."""

        self.iteration_editor.set_available_parameter_names(parameter_names)

    async def exec_async(self) -> None:
        while True:
            await self.watch_sequence()
            await anyio.sleep(10e-3)

    async def watch_sequence(self) -> None:
        if isinstance(self._state, _SequenceNotSetState):
            return
        elif isinstance(self._state, _SequenceSetState):
            old_state = self._state
            new_state = await _query_state_async(
                self._state.sequence_path, self.session_maker
            )
            # It could be that the widget state changed while we were fetching infos,
            # for example if set_sequence is called or the editors emitted editions.
            if old_state != self._state:
                return

            if new_state != self._state:
                self._transition(new_state)
        else:
            assert_never(self._state)

    def _transition(self, new_state: _State) -> None:
        match new_state:
            case _SequenceNotSetState():
                self.setVisible(False)
                self.start_sequence_action.setEnabled(False)
                self.interrupt_sequence_action.setEnabled(False)
                self.warning_action.setVisible(False)
            case _SequenceSetState(
                iterations=iterations,
                time_lanes=time_lanes,
                sequence_state=state,
                sequence_path=path,
            ):
                if not isinstance(iterations, StepsConfiguration):
                    raise NotImplementedError(f"Only supports {StepsConfiguration}")
                self.iteration_editor.set_iteration(iterations)
                self.time_lanes_editor.set_time_lanes(time_lanes)
                self._set_status_widget(path, state)
                if isinstance(new_state, _SequenceEditableState):
                    self.start_sequence_action.setEnabled(True)
                    self.interrupt_sequence_action.setEnabled(False)
                    self.time_lanes_editor.set_read_only(False)
                    self.iteration_editor.set_read_only(False)
                    self.tabWidget.setTabEnabled(0, False)
                    self.warning_action.setVisible(False)
                elif isinstance(new_state, _SequenceNotEditableState):
                    self.start_sequence_action.setEnabled(False)
                    self.interrupt_sequence_action.setEnabled(False)
                    self.time_lanes_editor.set_read_only(True)
                    self.iteration_editor.set_read_only(True)
                    self.parameters_editor.set_parameters(new_state.parameters)
                    self.tabWidget.setTabEnabled(0, True)
                    if isinstance(new_state, _SequenceCrashedState):
                        self.warning_action.setVisible(True)
                    else:
                        self.warning_action.setVisible(False)

                self.setVisible(True)
        self._state = new_state

    def _on_warning_action_triggered(self) -> None:
        """Display the sequence traceback in a dialog."""

        assert isinstance(self._state, _SequenceCrashedState)

        color = QApplication.palette().color(QPalette.ColorRole.Accent).name()
        self._exception_dialog.set_message(
            f"An error occurred while running the sequence "
            f"<b><font color='{color}'>{self._state.sequence_path}</font></b>."
        )
        traceback = self._state.exception_traceback
        self._exception_dialog.set_exception(traceback)
        self._exception_dialog.show()

    def _set_status_widget(self, path: PureSequencePath, state: State) -> None:
        text = " > ".join(path.parts)
        color = self.palette().text().color()
        if state.is_editable():
            icon = get_icon("editable-sequence", color=color)
        else:
            icon = get_icon("read-only-sequence", color=color)
        self.status_widget.set_text(text)
        self.status_widget.set_icon(icon)

    def _on_start_sequence_requested(self):
        assert isinstance(self._state, _SequenceEditableState)
        self.sequence_start_requested.emit(self._state.sequence_path)

    def set_sequence(self, sequence_path: Optional[PureSequencePath]) -> None:
        if sequence_path is None:
            new_state = _SequenceNotSetState()
        else:
            new_state = _query_state_sync(sequence_path, self.session_maker)
        self._transition(new_state)

    def setup_connections(self):
        self.time_lanes_editor.time_lanes_edited.connect(self.on_time_lanes_edited)
        self.iteration_editor.iteration_edited.connect(
            self.on_sequence_iteration_edited
        )

    def on_sequence_iteration_edited(self, iterations: IterationConfiguration):
        assert isinstance(self._state, _SequenceEditableState)
        with self.session_maker() as session:
            try:
                session.sequences.set_iteration_configuration(
                    self._state.sequence_path, iterations
                )
            except (PathNotFoundError, PathIsNotSequenceError):
                self._transition(_SequenceNotSetState())
            except SequenceNotEditableError:
                self._transition(
                    _query_state_sync(self._state.sequence_path, self.session_maker)
                )
            else:
                self._state = attrs.evolve(self._state, iterations=iterations)

    def on_time_lanes_edited(self, time_lanes: TimeLanes):
        assert isinstance(self._state, _SequenceEditableState)
        with self.session_maker() as session:
            try:
                session.sequences.set_time_lanes(self._state.sequence_path, time_lanes)
            except (PathNotFoundError, PathIsNotSequenceError):
                self._transition(_SequenceNotSetState())
            except SequenceNotEditableError:
                self._transition(
                    _query_state_sync(self._state.sequence_path, self.session_maker)
                )
            else:
                self._state = attrs.evolve(self._state, time_lanes=time_lanes)


class _State(abc.ABC):
    pass


@attrs.frozen
class _SequenceNotSetState(_State):
    pass


@attrs.frozen
class _SequenceSetState(_State):
    sequence_path: PureSequencePath
    sequence_state: State

    iterations: IterationConfiguration
    time_lanes: TimeLanes


@attrs.frozen
class _SequenceNotEditableState(_SequenceSetState):
    parameters: ParameterNamespace


@attrs.frozen
class _SequenceCrashedState(_SequenceNotEditableState):
    exception_traceback: Optional[TracebackSummary]


@attrs.frozen
class _SequenceEditableState(_SequenceSetState):
    pass


async def _query_state_async(
    path: PureSequencePath, session_maker: ExperimentSessionMaker
) -> _State:
    async with session_maker.async_session() as session:
        is_sequence_result = await session.sequences.is_sequence(path)
        try:
            is_sequence = unwrap(is_sequence_result)
        except PathNotFoundError:
            return _SequenceNotSetState()
        else:
            if is_sequence:
                return await _query_sequence_state_async(path, session)
            else:
                return _SequenceNotSetState()


async def _query_sequence_state_async(
    path: PureSequencePath, session: AsyncExperimentSession
) -> _SequenceSetState:
    # These results can be unwrapped safely because we checked that the sequence
    # exists in the session.
    state = unwrap(await session.sequences.get_state(path))
    iterations = await session.sequences.get_iteration_configuration(path)
    time_lanes = await session.sequences.get_time_lanes(path)

    if state.is_editable():
        return _SequenceEditableState(
            path, iterations=iterations, time_lanes=time_lanes, sequence_state=state
        )
    else:
        parameters = await session.sequences.get_global_parameters(path)
        if state == State.CRASHED:
            tb_summary = unwrap(await session.sequences.get_traceback_summary(path))
            return _SequenceCrashedState(
                path,
                iterations=iterations,
                time_lanes=time_lanes,
                parameters=parameters,
                sequence_state=state,
                exception_traceback=tb_summary,
            )
        else:
            return _SequenceNotEditableState(
                path,
                iterations=iterations,
                time_lanes=time_lanes,
                parameters=parameters,
                sequence_state=state,
            )


def _query_state_sync(
    path: PureSequencePath, session_maker: ExperimentSessionMaker
) -> _State:
    with session_maker() as session:
        is_sequence_result = session.sequences.is_sequence(path)
        try:
            is_sequence = unwrap(is_sequence_result)
        except PathNotFoundError:
            return _SequenceNotSetState()
        else:
            if is_sequence:
                return _query_sequence_state_sync(path, session)
            else:
                return _SequenceNotSetState()


def _query_sequence_state_sync(
    path: PureSequencePath, session: ExperimentSession
) -> _SequenceSetState:
    # These results can be unwrapped safely because we checked that the sequence
    # exists in the session.
    state = unwrap(session.sequences.get_state(path))
    iterations = session.sequences.get_iteration_configuration(path)
    time_lanes = session.sequences.get_time_lanes(path)

    if state.is_editable():
        return _SequenceEditableState(
            path, iterations=iterations, time_lanes=time_lanes, sequence_state=state
        )
    else:
        parameters = session.sequences.get_global_parameters(path)
        if state == State.CRASHED:
            traceback_summary = unwrap(session.sequences.get_exception(path))
            return _SequenceCrashedState(
                path,
                iterations=iterations,
                time_lanes=time_lanes,
                parameters=parameters,
                sequence_state=state,
                exception_traceback=traceback_summary,
            )
        else:
            return _SequenceNotEditableState(
                path,
                iterations=iterations,
                time_lanes=time_lanes,
                parameters=parameters,
                sequence_state=state,
            )


class IconLabel(QWidget):
    def __init__(
        self,
        parent: Optional[QWidget] = None,
        icon_position: Literal["left", "right"] = "left",
    ):
        super().__init__(parent)
        self._label = QLabel()
        self._icon = QLabel()
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        if icon_position == "left":
            layout.addWidget(self._icon)
            layout.addWidget(self._label)
        else:
            layout.addWidget(self._label)
            layout.addWidget(self._icon)
        self.setLayout(layout)

    def set_text(self, text: str):
        self._label.setText(text)

    def set_icon(self, icon: Optional[QIcon]):
        if icon is None:
            self._icon.clear()
        else:
            self._icon.setPixmap(icon.pixmap(20, 20))
